/* sane - Scanner Access Now Easy.

   Copyright (C) 2000 Mustek.
   Maintained by Tom Wang <tom.wang@mustek.com.tw>

   Updates (C) 2001 by Henning Meier-Geinitz.

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a SANE backend for Mustek 1200UB flatbed scanners.  */

#define BUILD 1

#include "../include/sane/config.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "../include/sane/sane.h"
#include "../include/sane/sanei.h"
#include "../include/sane/saneopts.h"
#include "mustek_usb.h"


#define BACKEND_NAME mustek_usb

#include "../include/sane/sanei_backend.h"
#include "../include/sane/sanei_config.h"

#include "mustek_usb_high.c"

static int num_devices;
static Mustek_Usb_Device *first_dev;
static Mustek_Usb_Scanner *first_handle;

static SANE_String_Const mode_list[6];

static SANE_String_Const SupportedModel[] =
{
"1",
"1200UB",
NULL
};

static const SANE_Int resbit_list[] =
{
  7,
  50, 100, 150, 200, 300, 400, 600
};

static const SANE_String_Const color_sensor_list[] =
{
  "All", "Red", "Green", "Blue",
  0
};

/* NOTE: This is used for Brightness, Contrast, Threshold, AutoBackAdj
   and 0 is the default value */
   
static const SANE_Range percentage_range =
{
	-100 << SANE_FIXED_SCALE_SHIFT,
	 100 << SANE_FIXED_SCALE_SHIFT,
	   0 << SANE_FIXED_SCALE_SHIFT
};

static const SANE_Range byte_range =
{
	1, 255, 1
};

static const SANE_Range char_range =
{
	-127, 127,1
};

static const SANE_Range u8_range =
{
	0,				/* minimum */
	255,				/* maximum */
	0				/* quantization */
};

static SANE_Bool 
CreateMappingTable (SANE_Word* pTable, SANE_Word inputLevel,
		    SANE_Word outputLevel, SANE_Int brightness,
		    SANE_Int contrast)
{	
  SANE_Word i;	
  SANE_Int inputBitsPerPixel = 0;	
  SANE_Int outputBitsPerPixel = 0;	
  SANE_Int temp;	
  double fOffset, ftemp;	
  SANE_Word iStart, iEnd, x0, y0;	
  double* lpCode;
  
  lpCode = (double*) malloc (inputLevel * sizeof (double));
  temp = inputLevel;	
  while (temp >= 2)
    {		
      temp = temp / 2;		
      inputBitsPerPixel ++;	
    }	
  temp = outputLevel;	
  while (temp >= 2)
    {		
      temp = temp / 2;		
      outputBitsPerPixel ++;	
    }	
  x0 = (SANE_Word) (inputLevel-1);	
  y0 = (SANE_Word) (outputLevel-1);	
  
  /* first, maps brightness, contrast effects to the table	
     Contrast first...	*/
  if (contrast >= 0)
    {		
      if (contrast > 0)			
	fOffset = (((double) (contrast) + 1.0f) * inputLevel / 256) - 1.0f;
      else			
	fOffset = 0.0f;		
      iStart = (SANE_Word) fOffset;		
      iEnd = (SANE_Word) (x0 - (SANE_Word) fOffset);		
      for (i=0; i<=x0; i++)
	{			
	  if (i<=iStart)				
	    lpCode[i] = 0;			
	  else if ((i > iStart) && (i < iEnd))				
	    lpCode[i] = (double) ((double)x0 * ( i - fOffset ) 
				  / ( x0 - 2 * fOffset ));
	  else if (i >= iEnd)				
	    lpCode[i] = (double) x0;		
	}	
    }	
  else
    {		
      fOffset = (((double) (contrast) - 1.0f) * inputLevel / (-256)) - 1.0f;
      iStart = 0;		
      iEnd = x0;		
      if (contrast == -127 )			
	for (i=0; i<=x0; i++)
	  lpCode[i] = (double) fOffset;		
      else			
	for (i=0; i<=x0; i++)
	  lpCode[i] = (double) (fOffset + (double) i * (x0 - 2 * fOffset)
				/ x0 ); 	
    }	
  /* apply brightness here */
  if (brightness > 0)		
    fOffset = ((double) (brightness+1) * inputLevel / 256) - 1;	
  else if (brightness < 0)		
    fOffset = ((double) (brightness-1) * inputLevel / 256) + 1;	
  else		
    fOffset = 0;	
  for (i=0; i<=x0; i++)
    {		
      ftemp = fOffset + lpCode[i];		
      if (ftemp <= 0)
	{			
	  lpCode[i] = 0;		
	} 
      else if (ftemp >= x0)
	{			
	  lpCode[i] = (double) x0;		
	} 
      else 			
	lpCode[i] = (double) ftemp;	
    }	/* end of brightness...	*/
		
  /* copy lpCode to pTable..	*/
  for (i = 0; i < inputLevel; i++)
    {		
      pTable[i] = (SANE_Word) lpCode[i];	
    }	
  free (lpCode);	
  /* map to pTable range to outputLevel	*/
  if (inputBitsPerPixel < outputBitsPerPixel)
    {
      SANE_Word wdTemp;
      wdTemp = 1;		
      wdTemp = wdTemp << (outputBitsPerPixel - inputBitsPerPixel - 1);
      for (i = 0 ; i < inputLevel; i++) 
	{
	  pTable[i] = (pTable[i] << (outputBitsPerPixel - inputBitsPerPixel))
	    + wdTemp;
	}
    } 	
  return SANE_TRUE;
}

static size_t 
max_string_size (const SANE_String_Const strings[])
{  
  size_t size, max_size = 0;  
  int i;  

  for (i = 0; strings[i]; ++i)
    {      
      size = strlen (strings[i]) + 1;      
      if (size > max_size)        
	max_size = size;    
    }  
  return max_size;
}

static SANE_Status
calc_parameters (Mustek_Usb_Scanner * s)
{
  SANE_String val = s->val[OPT_MODE].s;
  SANE_Status status = SANE_STATUS_GOOD;
  SANE_Bool  Protect = SANE_TRUE;
  SANE_Int max_x, max_y;

  DBG (5, "calc_parameters: start\n");
  
  if (!strcmp (val, "Lineart"))
    {
      s->params.last_frame = SANE_TRUE;
      s->params.format = SANE_FRAME_GRAY;
      s->params.depth = 1;
      s->nChannels = 1;
    }
  else if (!strcmp (val, "Gray"))
    {
      s->params.last_frame = SANE_TRUE;
      s->params.format = SANE_FRAME_GRAY;
      s->params.depth = 8;
      s->nChannels = 1;
    }
  else if (!strcmp (val, "Color"))
    {
      s->params.last_frame = SANE_TRUE;
      s->params.format = SANE_FRAME_RGB;
      s->params.depth = 8;
      s->nChannels = 3;
    }
  else
    {
      DBG (1, "calc_parameters: Invalid mode %s\n", (char *) val);
      status = SANE_STATUS_INVAL;
    }
  
  s->ulx = SANE_UNFIX (s->val[OPT_TL_X].w) / MM_PER_INCH;
  s->uly = SANE_UNFIX (s->val[OPT_TL_Y].w) / MM_PER_INCH;
  s->wx = SANE_UNFIX (s->val[OPT_BR_X].w) / MM_PER_INCH - s->ulx;
  s->wy = SANE_UNFIX (s->val[OPT_BR_Y].w) / MM_PER_INCH - s->uly;
  
  max_x = s->hw->MaxWidth*s->val[OPT_RESOLUTION].w/300;
  max_y = s->hw->MaxHeight*s->val[OPT_RESOLUTION].w/300;  
  
  s->ULx = s->ulx * s->val[OPT_RESOLUTION].w;
  s->Width = s->wx * s->val[OPT_RESOLUTION].w;
  s->ULy = s->uly * s->val[OPT_RESOLUTION].w;
  s->Height = s->wy * s->val[OPT_RESOLUTION].w;
  
  if (s->Width > max_x) s->Width = max_x;
  if (s->Height > max_y) s->Height = max_y;
  if (!strcmp (val, "Lineart")) {
    s->Width = (s->Width / 8) * 8;
    if (s->Width == 8) s->Width = 8;
  }
  if (s->ULx < 0) s->ULx = 0;
  if (s->ULy < 0) s->ULy = 0;
  if (s->ULx + s->Width > max_x) s->ULx = max_x - s->Width;
  if (s->ULy + s->Height> max_y) s->ULy = max_y - s->Height;
  
  if (Protect)
    {
      s->val[OPT_TL_X].w = SANE_FIX (s->ulx * MM_PER_INCH);
      s->val[OPT_TL_Y].w = SANE_FIX (s->uly * MM_PER_INCH);
      s->val[OPT_BR_X].w = SANE_FIX ((s->ulx + s->wx) * MM_PER_INCH);
      s->val[OPT_BR_Y].w = SANE_FIX ((s->uly + s->wy) * MM_PER_INCH);
    }
  else
    DBG (4, "calc_parameters: Not adapted. Protecting\n");
  
  s->params.pixels_per_line = s->Width;
  s->params.lines = s->Height;
  s->params.bytes_per_line = s->params.pixels_per_line * s->params.depth/8 
    * s->nChannels;
  
  DBG (4, "calc_parameters: format=%d\n", s->params.format);
  DBG (4, "calc_parameters: last_frame=%d\n", s->params.last_frame);
  DBG (4, "calc_parameters: lines=%d\n", s->params.lines);
  DBG (4, "calc_parameters: pixels_per_line=%d\n", 
       s->params.pixels_per_line);
  DBG (4, "calc_parameters: bytes_per_line=%d\n", s->params.bytes_per_line);
  DBG (4, "calc_parameters: Pixels %dx%dx%d\n",
       s->params.pixels_per_line, s->params.lines, 1 << s->params.depth);
  
  DBG (5, "calc_parameters: exit\n");
  return status;
}



static SANE_Status
init_options (Mustek_Usb_Scanner * s)
{
  int i;
  
  DBG (5, "init_options: start\n");

  memset (s->opt, 0, sizeof (s->opt));
  memset (s->val, 0, sizeof (s->val));
  
  for (i = 0; i < NUM_OPTIONS; ++i)
    {
      s->opt[i].size = sizeof (SANE_Word);
      s->opt[i].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    }

  s->opt[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].cap = SANE_CAP_SOFT_DETECT;
  s->val[OPT_NUM_OPTS].w = NUM_OPTIONS;

  /* "Mode" group: */
  s->opt[OPT_MODE_GROUP].title = "Scan Mode";
  s->opt[OPT_MODE_GROUP].desc = "";
  s->opt[OPT_MODE_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_MODE_GROUP].cap = 0;
  s->opt[OPT_MODE_GROUP].constraint_type = SANE_CONSTRAINT_NONE;
  
  mode_list[0]="Color";
  mode_list[1]="Gray";
  mode_list[2]="Lineart";
  mode_list[3]=NULL;
  
  /* scan mode */
  s->opt[OPT_MODE].name = SANE_NAME_SCAN_MODE;
  s->opt[OPT_MODE].title = SANE_TITLE_SCAN_MODE;
  s->opt[OPT_MODE].desc = SANE_DESC_SCAN_MODE;
  s->opt[OPT_MODE].type = SANE_TYPE_STRING;
  s->opt[OPT_MODE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
  s->opt[OPT_MODE].size = max_string_size (mode_list);
  s->opt[OPT_MODE].constraint.string_list = mode_list;
  s->val[OPT_MODE].s = strdup (mode_list[0]);
  
  /* resolution */
  s->opt[OPT_RESOLUTION].name = SANE_NAME_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].title = SANE_TITLE_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].desc = SANE_DESC_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].type = SANE_TYPE_INT;
  s->opt[OPT_RESOLUTION].unit = SANE_UNIT_DPI;
  /* TODO: Build the constraints on resolution in a smart way */
  s->opt[OPT_RESOLUTION].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_RESOLUTION].constraint.word_list = resbit_list;
  s->val[OPT_RESOLUTION].w = resbit_list[1];

  /* preview */
  s->opt[OPT_PREVIEW].name = SANE_NAME_PREVIEW;
  s->opt[OPT_PREVIEW].title = SANE_TITLE_PREVIEW;
  s->opt[OPT_PREVIEW].desc = SANE_DESC_PREVIEW;
  s->opt[OPT_PREVIEW].cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  s->opt[OPT_PREVIEW].type = SANE_TYPE_BOOL;
  s->val[OPT_PREVIEW].w = SANE_FALSE;

  /* "Geometry" group: */
  s->opt[OPT_GEOMETRY_GROUP].title = "Geometry";
  s->opt[OPT_GEOMETRY_GROUP].desc = "";
  s->opt[OPT_GEOMETRY_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_GEOMETRY_GROUP].cap = SANE_CAP_ADVANCED;
  s->opt[OPT_GEOMETRY_GROUP].constraint_type = SANE_CONSTRAINT_NONE;
  
  /* top-left x */
  s->opt[OPT_TL_X].name = SANE_NAME_SCAN_TL_X;
  s->opt[OPT_TL_X].title = SANE_TITLE_SCAN_TL_X;
  s->opt[OPT_TL_X].desc = SANE_DESC_SCAN_TL_X;
  s->opt[OPT_TL_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_X].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_X].constraint.range = &s->hw->x_range;
  s->val[OPT_TL_X].w = 0;
  
  /* top-left y */
  s->opt[OPT_TL_Y].name = SANE_NAME_SCAN_TL_Y;
  s->opt[OPT_TL_Y].title = SANE_TITLE_SCAN_TL_Y;
  s->opt[OPT_TL_Y].desc = SANE_DESC_SCAN_TL_Y;
  s->opt[OPT_TL_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_Y].constraint.range = &s->hw->y_range;
  s->val[OPT_TL_Y].w = 0;
  
  /* bottom-right x */
  s->opt[OPT_BR_X].name = SANE_NAME_SCAN_BR_X;
  s->opt[OPT_BR_X].title = SANE_TITLE_SCAN_BR_X;
  s->opt[OPT_BR_X].desc = SANE_DESC_SCAN_BR_X;
  s->opt[OPT_BR_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_X].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_X].constraint.range = &s->hw->x_range;
  s->val[OPT_BR_X].w = s->hw->x_range.max;
  
  /* bottom-right y */
  s->opt[OPT_BR_Y].name = SANE_NAME_SCAN_BR_Y;
  s->opt[OPT_BR_Y].title = SANE_TITLE_SCAN_BR_Y;
  s->opt[OPT_BR_Y].desc = SANE_DESC_SCAN_BR_Y;
  s->opt[OPT_BR_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_Y].constraint.range = &s->hw->y_range;
  s->val[OPT_BR_Y].w = s->hw->y_range.max;
  
  /* "Enhancement" group: */
  s->opt[OPT_ENHANCEMENT_GROUP].title = "Enhancement";
  s->opt[OPT_ENHANCEMENT_GROUP].desc = "";
  s->opt[OPT_ENHANCEMENT_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_ENHANCEMENT_GROUP].cap = 0;
  s->opt[OPT_ENHANCEMENT_GROUP].constraint_type = SANE_CONSTRAINT_NONE;
  
  /* brightness */
  s->opt[OPT_BRIGHTNESS].name = SANE_NAME_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].title = SANE_TITLE_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].desc = SANE_DESC_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].type = SANE_TYPE_INT;
  s->opt[OPT_BRIGHTNESS].unit = SANE_UNIT_NONE;
  s->opt[OPT_BRIGHTNESS].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BRIGHTNESS].constraint.range = &char_range;
  s->val[OPT_BRIGHTNESS].w = 0;
  
  /* contrast */
  s->opt[OPT_CONTRAST].name = SANE_NAME_CONTRAST;
  s->opt[OPT_CONTRAST].title = SANE_TITLE_CONTRAST;
  s->opt[OPT_CONTRAST].desc = SANE_DESC_CONTRAST
    " This option is active for halftone/Grayscale modes only.";
  s->opt[OPT_CONTRAST].type = SANE_TYPE_INT;
  s->opt[OPT_CONTRAST].unit = SANE_UNIT_NONE;
  s->opt[OPT_CONTRAST].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_CONTRAST].constraint.range = &char_range;
  s->val[OPT_CONTRAST].w = 0;
  
  /* custom-gamma table */
  s->opt[OPT_CUSTOM_GAMMA].name = SANE_NAME_CUSTOM_GAMMA;
  s->opt[OPT_CUSTOM_GAMMA].title = SANE_TITLE_CUSTOM_GAMMA;
  s->opt[OPT_CUSTOM_GAMMA].desc = SANE_DESC_CUSTOM_GAMMA;
  s->opt[OPT_CUSTOM_GAMMA].type = SANE_TYPE_BOOL;
  s->opt[OPT_CUSTOM_GAMMA].cap |= SANE_CAP_INACTIVE;
  s->val[OPT_CUSTOM_GAMMA].w = SANE_FALSE;
  
  /* red gamma vector */
  s->opt[OPT_GAMMA_VECTOR_R].name = SANE_NAME_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].title = SANE_TITLE_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].desc = SANE_DESC_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_R].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_R].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_R].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_R].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_R].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_R].wa = NULL;
  
  /* green gamma vector */
  s->opt[OPT_GAMMA_VECTOR_G].name = SANE_NAME_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].title = SANE_TITLE_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].desc = SANE_DESC_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_G].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_G].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_G].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_G].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_G].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_G].wa = NULL;
  
  /* blue gamma vector */
  s->opt[OPT_GAMMA_VECTOR_B].name = SANE_NAME_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].title = SANE_TITLE_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].desc = SANE_DESC_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_B].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_B].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_B].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_B].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_B].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_B].wa = NULL; /* &s->gamma_table[2][0]; */
  
  calc_parameters (s);

  DBG (5, "init_options: exit\n");
  return SANE_STATUS_GOOD;
}

static SANE_Status
attach (const char *devname, Mustek_Usb_Device ** devp, int may_wait)
{
  Mustek_Usb_Device *dev;
  FILE* fp;

  DBG (5, "attach: start: devp %s NULL, may_wait = %d\n", devp ? "!=" : "==", 
       may_wait);
  if (!devname)
    {
      DBG (1, "attach: devname == NULL\n");
      return SANE_STATUS_INVAL;
    }

  for (dev = first_dev; dev; dev = dev->next)
    if (strcmp (dev->sane.name, devname) == 0)
      {
	if (devp)
	  *devp = dev;
	DBG (4, "attach: device %s was already in device list\n", devname);
	return SANE_STATUS_GOOD;
      }

  dev = malloc (sizeof (Mustek_Usb_Device));
  if (!dev)
    {
      DBG(1, "attach: couldn't malloc Mustek_Usb_Device\n");
      return SANE_STATUS_NO_MEM;
    }

  /* check if the device is present. */
  DBG(4, "attach: trying to open %s\n", devname);
  fp = fopen(devname, "r+w");
  if (!fp) 
    {
      DBG(1, "attach: couldn't open %s: %s\n", devname, strerror (errno));
      return SANE_STATUS_GOOD;
    }
  else
    DBG(3, "attach: %s seems to exist\n", devname);
    
  fclose(fp);
  	
  memset(dev, 0, sizeof (*dev));
  dev->name = strdup (devname);
  dev->sane.name = (SANE_String_Const) dev->name;
  dev->sane.vendor = "Mustek";
  dev->sane.model = "1200 UB";
  dev->sane.type = "flatbed scanner";

  dev->x_range.min = 0;
  dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);
  dev->x_range.quant = 0;

  dev->y_range.min = 0;
  dev->y_range.max = SANE_FIX (11.9 * MM_PER_INCH);
  dev->y_range.quant = 0;

  dev->MaxHeight = 11.9*300;
  dev->MaxWidth = 8.5*300;
  dev->dpi_range.min = SANE_FIX (50);
  dev->dpi_range.max = SANE_FIX (600);
  dev->dpi_range.quant = SANE_FIX (1);

  if (!usb_high_scan_init (dev))
    {
      DBG (1, "attach: usb_high_scan_init returned SANE_FALSE\n");
      return SANE_STATUS_INVAL;
    }

  DBG (2, "attach: found %s %s %s at %s\n", dev->sane.vendor, dev->sane.type,
       dev->sane.model, dev->sane.name);
  ++num_devices;
  dev->next = first_dev;
  first_dev = dev;

  if (devp)
    *devp = dev;

  DBG (5, "attach: exit\n");
  return SANE_STATUS_GOOD;
}

static void 
SetAndGetParameters (Mustek_Usb_Scanner* s)
{
  TARGETIMAGE tiTarget;
  SANE_String val;

  val = s->val[OPT_MODE].s;
  
  if (!strcmp (val, "Lineart"))
    {
      tiTarget.cmColorMode = CGRAY8;
    }
  else if (!strcmp (val, "Gray"))
    {
      tiTarget.cmColorMode = CGRAY8;
    }
  else /* Color */
    {
      tiTarget.cmColorMode = CRGB24;
    }
  
  tiTarget.isOptimalSpeed = SANE_TRUE;
  tiTarget.wDpi = s->val[OPT_RESOLUTION].w;
  tiTarget.wX = (SANE_Word)(s->ULx);
  tiTarget.wY = (SANE_Word)(s->ULy);
  tiTarget.wWidth = (SANE_Word)(s->Width);
  tiTarget.wHeight = (SANE_Word)(s->Height);
  if (!s->hw->Prepared)
    {
      usb_high_scan_Prepare (s->hw);
      usb_high_scan_Reset (s->hw);
      s->hw->Prepared = SANE_TRUE;
    }
  usb_high_scan_SetThreshold (s->hw, 128);
  usb_high_scan_EmbedGamma (s->hw, NULL);
  usb_high_scan_ScanSuggest (s->hw, &tiTarget, &s->hw->Suggest);	
}

static void 
StartScan (Mustek_Usb_Scanner *s)
{
  DBG (5, "StartScan: s=%p\n", s);
  usb_high_scan_SetupScan (s->hw, s->hw->Suggest.cmScanMode,
			   s->hw->Suggest.wXDpi, s->hw->Suggest.wYDpi, 0,
			   s->hw->Suggest.wX, s->hw->Suggest.wY,
			   s->hw->Suggest.wWidth);
}

/* -------------------------- SANE API functions ------------------------- */

SANE_Status
sane_init (SANE_Int * version_code, SANE_Auth_Callback authorize)
{
  char dev_name[PATH_MAX];
  size_t len;
  FILE *fp;

  DBG_INIT ();
  DBG (2, "SANE Mustek USB backend version %d.%d build %d from %s\n", V_MAJOR,
       V_MINOR, BUILD, PACKAGE_VERSION);

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, BUILD);

  DBG(5, "sane_init: authorize %s null\n", authorize ? "!=" : "==");

  fp = sanei_config_open (MUSTEK_USB_CONFIG_FILE);
  if (!fp)
    {
      /* default to /dev/usb/scanner instead of insisting on config file */
      DBG(3, "sane_init: couldn't open config file `%s: %s. Using "
	  "/dev/usb/scanner directly\n", MUSTEK_USB_CONFIG_FILE,
	  strerror (errno));
      attach ("/dev/usb/scanner", 0, SANE_FALSE);
      return SANE_STATUS_GOOD;
    }

  while (sanei_config_read (dev_name, sizeof (dev_name), fp))
    {
      if (dev_name[0] == '#')	/* ignore line comments */
	continue;

      len = strlen (dev_name);

      if (!len)
	continue;		/* ignore empty lines */

      if (strncmp (dev_name, "option", 6) == 0
	  && isspace (dev_name[6]))
	{
	  const char *str = dev_name + 7;

	  while (isspace (*str))
	    ++str;
	  
	  continue;
	}
      attach (dev_name, 0, SANE_FALSE);
    }
  fclose (fp);
  DBG(5, "sane_init: exit\n");

  return SANE_STATUS_GOOD;
}

void
sane_exit (void)
{
  Mustek_Usb_Device *dev, *next;
  
  DBG(5, "sane_exit: start\n");
  for (dev = first_dev; dev; dev = next)
    {
      next = dev->next;
      usb_high_scan_exit (dev);
      free ((void *) dev->name);
      free (dev);
    }
  DBG(5, "sane_exit: exit\n");
}

SANE_Status
sane_get_devices (const SANE_Device *** device_list, SANE_Bool local_only)
{
  static const SANE_Device **devlist = 0;
  Mustek_Usb_Device *dev;
  int i;
  
  DBG(5, "sane_get_devices: start: local_only = %s\n", 
      local_only == SANE_TRUE ? "true" : "false");

  if (devlist)
    free (devlist);
  
  devlist = malloc ((num_devices + 1) * sizeof (devlist[0]));
  if (!devlist)
    return SANE_STATUS_NO_MEM;
  
  i = 0;
  for (dev = first_dev; i < num_devices; dev = dev->next) 
    {
      /* check if the scanner is available. */
      FILE* fp;
      fp = fopen(dev->sane.name,"r+w");
      if (fp)
	{
	  DBG(5, "sane_get_devices: %s could be opened\n", dev->sane.name);
	  devlist[i++] = &dev->sane;
	  fclose(fp);
	} 
      else
	{
	  DBG(5, "sane_get_devices: %s could not be opened\n", dev->sane.name);
	  i++;
	}
    }
  devlist[i++] = 0;
  
  *device_list = devlist;

  DBG(5, "sane_get_devices: exit\n");

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const devicename, SANE_Handle * handle)
{
  Mustek_Usb_Device *dev;
  SANE_Status status;
  Mustek_Usb_Scanner *s;
  
  DBG(5, "sane_open: start\n");

  if (devicename[0])
    {
      for (dev = first_dev; dev; dev = dev->next)
	if (strcmp (dev->sane.name, devicename) == 0) 
	  break;
      
      if (!dev)
	{
	  status = attach (devicename, &dev, SANE_TRUE);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}
    }
  else
    /* empty devicname -> use first device */
    dev = first_dev;
  
  if (!dev)
    return SANE_STATUS_INVAL;
  
  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;
  memset (s, 0, sizeof (*s));
  s->hw = dev;
  
  init_options (s);
  
  /* insert newly opened handle into list of open handles: */
  s->next = first_handle;
  first_handle = s;
  
  *handle = s;
  usb_high_scan_Init (s->hw, 1, devicename);
  usb_high_scan_TurnPower (s->hw, SANE_TRUE);
  usb_high_scan_BackHome (s->hw);

  DBG(5, "sane_open: exit\n");

  return SANE_STATUS_GOOD;
}

void
sane_close (SANE_Handle handle)
{
  Mustek_Usb_Scanner *prev, *s;

  DBG(5, "sane_close: start\n");

  /* remove handle from list of open handles: */
  prev = 0;
  for (s = first_handle; s; s = s->next)
    {
      if (s == handle)
	break;
      prev = s;
    }
  if (!s)
    {
      DBG (5, "close: invalid handle %p\n", handle);
      return;			/* oops, not a handle we know about */
    }
  
  if (prev)
    prev->next = s->next;
  else
    first_handle = s->next;
  
  if (s->hw->Prepared)
    usb_high_scan_Clearup (s->hw);
  usb_high_scan_Release (s->hw);

  free (handle);

  DBG(5, "sane_close: exit\n");
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  Mustek_Usb_Scanner *s = handle;
  
  if ((unsigned) option >= NUM_OPTIONS)
    return 0;
  return s->opt + option;
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *val, SANE_Int * info)
{
  Mustek_Usb_Scanner *s = handle;
  SANE_Status status;
  SANE_Word cap;

  DBG (5, "sane_control_option: start: action = %s, option = %s (%d)\n",
       (action == SANE_ACTION_GET_VALUE) ? "get" : "set",
       s->opt[option].name, option);
  
  if (val || action == SANE_ACTION_GET_VALUE)
    switch (s->opt[option].type)
      {
      case SANE_TYPE_STRING:
	DBG (5, "sane_control_option: Value %s\n", s->val[option].s);
	break;
      case SANE_TYPE_FIXED:
	{
	  double v1, v2;
	  SANE_Fixed f;
	  v1 = SANE_UNFIX (s->val[option].w);
	  f = *(SANE_Fixed *) val;
	  v2 = SANE_UNFIX (f);
	  DBG (5, "sane_control_option: Value %g (Fixed)\n", v1);
	}
	break;
      default:
	DBG (5, "Value %u (Int)\n", s->val[option].w);
	break;
      }
  
  if (info)
    *info = 0;
  
  if (s->scanning)
    return SANE_STATUS_DEVICE_BUSY;
  
  if (option >= NUM_OPTIONS)
    return SANE_STATUS_INVAL;
  
  cap = s->opt[option].cap;
  
  if (!SANE_OPTION_IS_ACTIVE (cap))
    return SANE_STATUS_INVAL;
  
  if (action == SANE_ACTION_GET_VALUE)
    {
      switch (option)
	{
	  /* word options: */
	case OPT_NUM_OPTS:
	case OPT_RESOLUTION:
	case OPT_PREVIEW:
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	case OPT_BRIGHTNESS:
	case OPT_CONTRAST:
	case OPT_CUSTOM_GAMMA:
	  *(SANE_Word *) val = s->val[option].w;
	  return SANE_STATUS_GOOD;
	  
	  /* word-array options: */
	case OPT_GAMMA_VECTOR_R:
	case OPT_GAMMA_VECTOR_G:
	case OPT_GAMMA_VECTOR_B:
	  memcpy (val, s->val[option].wa, s->opt[option].size);
	  return SANE_STATUS_GOOD;
	  
	  /* string options: */
	case OPT_MODE:
	  strcpy(val, s->val[option].s);
	  return SANE_STATUS_GOOD;
	}
    }
  else if (action == SANE_ACTION_SET_VALUE)
    {
      if (!SANE_OPTION_IS_SETTABLE (cap))
	return SANE_STATUS_INVAL;
      
      status = sanei_constrain_value (s->opt + option, val, info);
      
      if (status != SANE_STATUS_GOOD)
	return status;
      
      switch (option)
	{
	  /* (mostly) side-effect-free word options: */
	case OPT_RESOLUTION:
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	  s->val[option].w = *(SANE_Word *) val;
	  calc_parameters (s);
	  if (info)
	    *info |= SANE_INFO_INEXACT|SANE_INFO_RELOAD_PARAMS;
	  return SANE_STATUS_GOOD;
	  
	  /* fall through */
	case OPT_BRIGHTNESS:
	case OPT_CONTRAST:
	  s->val[option].w = *(SANE_Word *) val;
	  return SANE_STATUS_GOOD;
	  /* Boolean */
	case OPT_PREVIEW:
	  s->val[option].w = *(SANE_Bool *) val;
	  return SANE_STATUS_GOOD;
	  
	  /* side-effect-free word-array options: */
	case OPT_GAMMA_VECTOR_R:
	case OPT_GAMMA_VECTOR_G:
	case OPT_GAMMA_VECTOR_B:
	  memcpy (s->val[option].wa, val, s->opt[option].size);
	  return SANE_STATUS_GOOD;
	  
	case OPT_CUSTOM_GAMMA:
	  s->val[OPT_CUSTOM_GAMMA].w = *(SANE_Word *) val;
	  /* ??? gamma_update(s); */
	  if (info)
	    *info |= SANE_INFO_RELOAD_OPTIONS;
	  return SANE_STATUS_GOOD;
	  
	case OPT_MODE:
	  if (s->val[option].s)
	    free (s->val[option].s);
	  s->val[option].s = strdup (val);
	  
	  status = calc_parameters(s);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	  if (info)
	    *info |= SANE_INFO_RELOAD_OPTIONS | SANE_INFO_RELOAD_PARAMS;
	  return SANE_STATUS_GOOD;
	} /* End of switch */
    } /* End of SET_VALUE */

  DBG(4, "sane_control_option: unknown action for option %s\n",
      s->opt[option].name);
  
  return SANE_STATUS_INVAL;
}

SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  Mustek_Usb_Scanner *s = handle;
  
  DBG(5, "sane_get_parameters: start\n");

  calc_parameters (s);
  if (params)
    *params = s->params;

  DBG(5, "sane_get_parameters: exit\n");

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_start (SANE_Handle handle)
{
  Mustek_Usb_Scanner *s = handle;
  
  DBG(5, "sane_start: start\n");

  /* First make sure we have a current parameter set.  Some of the
     parameters will be overwritten below, but that's OK.  */
  
  calc_parameters (s);
  CreateMappingTable (s->hw->mapping_table, 256, 256, 
		      s->val[OPT_BRIGHTNESS].w,
		      s->val[OPT_CONTRAST].w);
  SetAndGetParameters (s);
  StartScan (s);
  s->read_rows = s->Height; 

  DBG(5, "sane_start: exit\n");

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * buf, SANE_Int max_len,
	   SANE_Int * len)
{
  Mustek_Usb_Scanner *s = handle;
  int lines_read;
  int i,j;
  SANE_String val = s->val[OPT_MODE].s;
  int bpp;
  int bright, contra;
  
  DBG(5, "sane_read: start\n");

  bright = s->val[OPT_BRIGHTNESS].w;
  contra = s->val[OPT_CONTRAST].w;
  
  if (!strcmp (val, "Lineart"))
    {
      bpp = 8;
    }
  else if (!strcmp (val, "Gray"))
    {
      bpp = 8;
    }
  else /* Color */
    {
      bpp = 24;
    }
  
  *len = 0;
  if (s->read_rows > 0)
    {
      /* We don't need to BackTrack... too waste time.
	 if (a->read_rows != s->Height) CScanner__BackTrack(&m_Scanner);
      */
      lines_read = max_len / (s->Width*bpp/8);
      if (lines_read > s->read_rows)
	lines_read = s->read_rows;
      usb_high_scan_GetRows (s->hw, buf, lines_read, SANE_FALSE);
    if (!strcmp (val, "Lineart"))
      {
	long int pointer;
	pointer = 0;
	bright = 128 - bright;
	for (i = 0; i < lines_read; i++)
	  for (j = 0; j<s->Width/8; j++) 
	    {
	      buf[pointer] = 
		(((buf[i * s->Width + j * 8 + 0] > bright) ? 0 : 1) << 7) +
		(((buf[i * s->Width + j * 8 + 1] > bright) ? 0 : 1) << 6) +
		(((buf[i * s->Width + j * 8 + 2] > bright) ? 0 : 1) << 5) +
		(((buf[i * s->Width + j * 8 + 3] > bright) ? 0 : 1) << 4) +
		(((buf[i * s->Width + j * 8 + 4] > bright) ? 0 : 1) << 3) +
		(((buf[i * s->Width + j * 8 + 5] > bright) ? 0 : 1) << 2) +
		(((buf[i * s->Width + j * 8 + 6] > bright) ? 0 : 1) << 1) +
		(((buf[i * s->Width + j * 8 + 7] > bright) ? 0 : 1));
	    pointer++;
	  }
	*len = pointer;
	s->read_rows -= lines_read;
      } 
    else
      {
	*len = (s->Width*bpp/8)*lines_read;
	s->read_rows -= lines_read;
	for (i = 0; i < *len; i++)
	  {
	    buf[i] = s->hw->mapping_table[buf[i]];
	  }
      }
    } 
  else
    {
      return SANE_STATUS_EOF;
    }
  DBG(5, "sane_read: exit\n");
  return SANE_STATUS_GOOD;
}

void
sane_cancel (SANE_Handle handle)
{
  Mustek_Usb_Scanner *s = handle;
  
  DBG (5, "sane_cancel: start\n");

  usb_high_scan_StopScan (s->hw);
  usb_high_scan_BackHome (s->hw);

  if (s->scanning)
    {
      if (s->aborted_by_user)
	{
	  DBG (3,"sane_cancel: Allready Aborted. Please Wait...\n");
	}
      else
	{
	  s->scanning = SANE_FALSE;
	  s->aborted_by_user = SANE_TRUE;
	  DBG (3, "sane_cancel: Signal Caught! Aborting...\n");
	}
    }
  else
    {
      if (s->aborted_by_user)
	{
	  DBG (3, "sane_cancel: Scan has not been Initiated yet, "
	       "or it is allready aborted.\n");
	  s->aborted_by_user = SANE_FALSE;
	}
      else
	{
	  DBG (3, "sane_cancel: Scan has not been Initiated "
	       "yet (or it's over).\n");
	}
    }
  return;
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  DBG (5, "sane_set_io_mode: handle = %p, non_blockeing = %s\n", 
       handle, non_blocking == SANE_TRUE ? "true" : "false");
  return SANE_STATUS_UNSUPPORTED;
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  DBG (5, "sane_get_select_fd: handle = %p, fd = %p\n", handle, fd);
  return SANE_STATUS_UNSUPPORTED;
}
