/* sane - Scanner Access Now Easy.

   Copyright (C) 2000 Mustek.
   Maintained by Tom Wang <tom.wang@mustek.com.tw>

   Updates (C) 2001 by Henning Meier-Geinitz.

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a SANE backend for Mustek 1200UB flatbed scanners.  */

#include "mustek_usb_high.h"
#include "mustek_usb_mid.c"

/* calibrator ------------------------------------------------------------ */

static SANE_Byte GrayMap[8] = 
{
  0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static inline double
FiltLower (int *pSortBuffer, SANE_Word TotalCount, SANE_Word FiltCount)
{
  SANE_Word Bound = TotalCount - 1;
  SANE_Word LeftCount = TotalCount - FiltCount;
  int Temp = 0;
  SANE_Word i, j;
  int Sum = 0;
  for (i = 0; i < Bound; i++)
    {
      for (j = 0; j < Bound - i; j++)
	{
	  if (pSortBuffer[j + 1] > pSortBuffer[j])
	    {
	      Temp = pSortBuffer[j];
	      pSortBuffer[j] = pSortBuffer[j + 1];
	      pSortBuffer[j + 1] = Temp;
	    }
	}
    }
  for (i = 0; i < LeftCount; i++)
    Sum += pSortBuffer[i];
  return (double) Sum;
}

void
usb_high_cal_MD12_init (CalMD12 * cal, SANE_Byte type)
{
  cal->isPrepared = SANE_FALSE;
  cal->pKWhite = NULL;
  cal->pKDark = NULL;
  /* Working Buffer */
  cal->pWhiteLine = NULL;
  cal->pDarkLine = NULL;
  cal->pWhiteBuffer = NULL;
  /* Necessary Parameters */
  cal->wKWhiteLevel = 240 << 8;
  cal->wKDarkLevel = 0;
  cal->wThreshold = 2048;
  cal->wMajorAver = 0;
  cal->wMinorAver = 0;
  cal->wFilter = 0;
  cal->wWhiteNeeded = 0;
  cal->wDarkNeeded = 0;
  cal->wMaxWidth = 0;
  cal->wWidth = 100;
  cal->pGammaTable = NULL;
  cal->nCalibratorType = type;
}

void
usb_high_cal_MD12_exit (CalMD12 * cal)
{
  if (cal->isPrepared)
    usb_high_cal_MD12_Clearup (cal);
}

void
usb_high_cal_MD12_Init (CalMD12 * cal, SANE_Word wTargetWhite,
		       SANE_Word wTargetDark)
{
  DBG(5, "usb_high_cal_MD12_Init: wTargetDark = %d\n", wTargetDark);
  cal->wKWhiteLevel = wTargetWhite / 16;
  cal->wKDarkLevel = 0;
}

void
usb_high_cal_MD12_Release (CalMD12 * cal)
{
  usb_high_cal_MD12_exit (cal);
}

void
usb_high_cal_MD12_SetThreshold (CalMD12 * cal, SANE_Word threshold)
{
  cal->wThreshold = threshold >> 4;
}

void
usb_high_cal_MD12_EmbedGamma (CalMD12 * cal, SANE_Word * pGammaTable)
{
  cal->pGammaTable = pGammaTable;
}

SANE_Bool usb_high_cal_MD12_Prepare (CalMD12 * cal, SANE_Word wMaxWidth)
{

  if (cal->isPrepared)
    {
      DBG (5, "usb_high_cal_MD12_Parepare isPrepared");
      return SANE_FALSE;
    }

  if (cal->pKWhite)
    {
      free (cal->pKWhite);
    }
  cal->pKWhite = (SANE_Word *) malloc (wMaxWidth * sizeof (SANE_Word));
  if (cal->pKDark)
    {
      free (cal->pKDark);
    }
  cal->pKDark = (SANE_Word *) malloc (wMaxWidth * sizeof (SANE_Word));
  cal->wMaxWidth = wMaxWidth;

  cal->isPrepared = SANE_TRUE;

  return SANE_TRUE;
}

SANE_Bool usb_high_cal_MD12_Clearup (CalMD12 * cal)
{
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12_Clearup !isPrepared");
      return SANE_FALSE;
    }

  if (cal->pKDark)
    {
      free (cal->pKDark);
    }
  cal->pKDark = NULL;
  if (cal->pKWhite)
    {
      free (cal->pKWhite);
    }
  cal->pKWhite = NULL;

  cal->isPrepared = SANE_FALSE;
  return SANE_TRUE;
}

#if 0
SANE_Bool usb_high_cal_MD12_Reset (CalMD12 * cal)
{
  SANE_Word i;
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12_Reset !isPrepared");
      return SANE_FALSE;
    }

  for (i = 0; i < cal->wMaxWidth; i++)
    {
      cal->pKWhite[i] = cal->wKWhiteLevel;
      cal->pKDark[i] = cal->wKDarkLevel;
    }
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_high_cal_MD12_Setup (CalMD12 * cal, SANE_Word wMojorAver,
			SANE_Word wMinorAver, SANE_Word wFilter, SANE_Word wWidth,
			SANE_Word * wWhiteNeeded, SANE_Word * wDarkNeeded)
{
  int i;
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12_Setup !isPrepared");
      return SANE_FALSE;
    }
  if (wMojorAver == 0)
    {
      DBG (5,"usb_high_cal_MD12_Setup wMojorAver==0");
      return SANE_FALSE;
    }
  if (wMinorAver == 0)
    {
      DBG (5,"usb_high_cal_MD12_Setup wMinorAver==0");
      return SANE_FALSE;
    }
  if (wWidth > cal->wMaxWidth)
    {
      DBG (5,"usb_high_cal_MD12_Setup wWidth>wMaxWidth");
      return SANE_FALSE;
    }

  cal->wMajorAver = wMojorAver;
  cal->wMinorAver = wMinorAver;
  cal->wFilter = wFilter;
  cal->wWidth = wWidth;
  cal->wWhiteNeeded = wMojorAver * 16 + wFilter;
  cal->wDarkNeeded = wMojorAver * 16;
  *wWhiteNeeded = cal->wWhiteNeeded;
  *wDarkNeeded = cal->wDarkNeeded;

  if (cal->pWhiteLine)
    {
      free (cal->pWhiteLine);
    }
  cal->pWhiteLine =
    (double *) malloc (cal->wWidth * sizeof (double));
  if (cal->pDarkLine)
    {
      free (cal->pDarkLine);
    }
  cal->pDarkLine =
    (double *) malloc (cal->wWidth * sizeof (double));
  for (i = 0; i < cal->wWidth; i++)
    {
      cal->pWhiteLine[i] = 0.0;
      cal->pDarkLine[i] = 0.0;
    }
  if (cal->pWhiteBuffer)
    {
      free (cal->pWhiteBuffer);
    }
  cal->pWhiteBuffer =
    (int *) malloc (cal->wWhiteNeeded * cal->wWidth * sizeof (int));

  for (i = 0; i < cal->wWhiteNeeded * cal->wWidth; i++)
    {
      *(cal->pWhiteBuffer + i) = 0;
    }
  return SANE_TRUE;
}

SANE_Bool usb_high_cal_MD12_EvalueWhite (CalMD12 * cal, double dFactor)
{
  /* Caculate pWhiteLine */
  double LoopDivision = (double) (cal->wMajorAver * cal->wMinorAver);
  double Aver;
  int *pSortBuffer = (int *) malloc (cal->wWhiteNeeded * sizeof (int));
  SANE_Word i, j;

  if (cal->pWhiteBuffer == NULL)
    {
      DBG (5,"usb_high_cal_MD12_EvalueWhite pWhiteBuffer==NULL");
      return SANE_FALSE;
    }

  for (i = 0; i < cal->wWidth; i++)
    {
      for (j = 0; j < cal->wWhiteNeeded; j++)
	{
	  *(pSortBuffer + j) =
	    *(cal->pWhiteBuffer + j * cal->wWidth + i);
	}
      Aver =
	FiltLower (pSortBuffer, cal->wWhiteNeeded,
		   cal->wFilter) * dFactor / LoopDivision;
      if (Aver >= 4096.0)
	cal->pWhiteLine[i] = 4095.9999;
      else if (Aver < 0.0)
	cal->pWhiteLine[i] = 0.0;
      else
	cal->pWhiteLine[i] = Aver;
    }
  free (pSortBuffer);
  pSortBuffer = NULL;
  free (cal->pWhiteBuffer);
  cal->pWhiteBuffer = NULL;
  return SANE_TRUE;
}

void
usb_high_cal_MD12_EvalueDark (CalMD12 * cal, double dFactor)
{
  SANE_Word i;
  double LoopDivision;
  /* Caculate pDarkLine */
  dFactor *= 16.0;
  LoopDivision = (double) (cal->wMajorAver * cal->wMinorAver);
  for (i = 0; i < cal->wWidth; i++)
    {
      cal->pDarkLine[i] /= LoopDivision;
      cal->pDarkLine[i] -= dFactor;
      if (cal->pDarkLine[i] < 0.0)
	cal->pDarkLine[i] = 0.0;
    }
}

SANE_Bool usb_high_cal_MD12_EvalueCalibrator (CalMD12 * cal)
{
  int Aver = 0;
  SANE_Word i;
  /*    FILE* pfile; */
  if (cal->pWhiteLine == NULL)
    {
      DBG (5,"usb_high_cal_MD12_EvalueCalibrator pWhiteLine==NULL");
      return SANE_FALSE;
    }
  if (cal->pDarkLine == NULL)
    {
      DBG (5,"usb_high_cal_MD12_EvalueCalibrator pDarkLine==NULL");
      return SANE_FALSE;
    }

  /*    pfile = fopen("/root/calibrator.log","a+w"); */
  for (i = 0; i < cal->wWidth; i++)
    {
      /*          char buf[256]; */
      Aver = (int) (cal->pWhiteLine[i]) - (int) (cal->pDarkLine[i]);
      if (Aver <= 0)
	Aver = 1;
      else if (Aver >= 4096)
	Aver = 4095;
      cal->pKWhite[i] = (SANE_Word) (Aver);
      cal->pKDark[i] = (SANE_Word) (cal->pDarkLine[i]);
      /*          sprintf(buf,"[%5d]: white = %4d, dark = %4d\n",i,Aver,cal->pKDark[i]); */
      /*          fwrite(buf,strlen(buf),1,pfile); */
    }
  /*    fclose(pfile); */
  free (cal->pDarkLine);
  cal->pDarkLine = NULL;
  free (cal->pWhiteLine);
  cal->pWhiteLine = NULL;

  return SANE_TRUE;
}

/* virtual function switcher */
SANE_Bool
usb_high_cal_MD12_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
			      SANE_Word wMinor, void * pWhitePattern)
{
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
    case I8O8MONO:
      return usb_high_cal_MD12i8o8_FillInWhite (cal, wMajor, wMinor,
					       pWhitePattern);
      break;
    case I4O1MONO:
      return usb_high_cal_MD12i4o1_FillInWhite (cal, wMajor, wMinor,
					       pWhitePattern);
      break;
    }
  return SANE_TRUE;
}

SANE_Bool
usb_high_cal_MD12_FillInDark (CalMD12 * cal, SANE_Word wMajor,
			     SANE_Word wMinor, void * pDarkPattern)
{
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
    case I8O8MONO:
      return usb_high_cal_MD12i8o8_FillInDark (cal, wMajor, wMinor,
					      pDarkPattern);
      break;
    case I4O1MONO:
      return usb_high_cal_MD12i4o1_FillInDark (cal, wMajor, wMinor,
					      pDarkPattern);
      break;
    }
  return SANE_TRUE;
}

void
usb_high_cal_MD12_Calibrate (CalMD12 * cal, void * pSrc,
			    void * pTarget)
{
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
      usb_high_cal_MD12i8o8Rgb_Calibrate (cal, pSrc, pTarget);
      break;
    case I8O8MONO:
      usb_high_cal_MD12i8o8Mono_Calibrate (cal, pSrc, pTarget);
      break;
    case I4O1MONO:
      usb_high_cal_MD12i4o1_Calibrate (cal, pSrc, pTarget);
      break;
    }
}

SANE_Bool
usb_high_cal_MD12i8o8_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
				  SANE_Word Minor, void * pWhitePattern)
{
  SANE_Byte * pPattern = (SANE_Byte *) pWhitePattern;
  SANE_Word j;

  DBG (5, "usb_high_cal_MD12i8o8_FillInWhite: Minor=%d\n", Minor);
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12i8o8_FillInWhite !isPrepared");
      return SANE_FALSE;
    }
  if (cal->wWhiteNeeded == 0)
    {
      DBG (5,"usb_high_cal_MD12i8o8_FillInWhite wWhiteNeeded==0");
      return SANE_FALSE;
    }

  for (j = 0; j < cal->wWidth; j++)
    {
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) (pPattern[j]);
    }
  return SANE_TRUE;
}

SANE_Bool
usb_high_cal_MD12i8o8_FillInDark (CalMD12 * cal, SANE_Word wMajor,
				  SANE_Word Minor, void * pDarkPattern)
{
  SANE_Byte * pPattern = (SANE_Byte *) pDarkPattern;
  SANE_Word j;

  DBG (5, "usb_high_cal_MD12i8o8FillInDark: wMajor=%d, Minor=%d\n",
       wMajor, Minor);
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12i8o8_FillInDark !isPrepared");
      return SANE_FALSE;
    }
  if (cal->wDarkNeeded == 0)
    {
      DBG (5,"usb_high_cal_MD12i8o8_FillInDark wDarkNeeded==0");
      return SANE_FALSE;
    }

  for (j = 0; j < cal->wWidth; j++)
    {
      cal->pDarkLine[j] += (double) (pPattern[j]);
    }
  return SANE_TRUE;
}

SANE_Bool
usb_high_cal_MD12i4o1_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
				   SANE_Word Minor, void * pWhitePattern)
{
  SANE_Byte * pPattern = (SANE_Byte *) pWhitePattern;
  SANE_Word j = 0;

  DBG(5, "usb_high_cal_MD12i4o1: Minor=%d\n", Minor);
  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12i4o1_FillInWhite !isPrepared");
      return SANE_FALSE;
    }
  if (cal->wWhiteNeeded == 0)
    {
      DBG (3, "usb_high_cal_MD12i4o1_FillInWhite wWhiteNeeded==0");
      return SANE_FALSE;
    }

  while (j < cal->wWidth)
    {
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) (*(pPattern) & 0xf0);
      j++;
      if (j >= cal->wWidth)
	break;
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) ((SANE_Byte) (*(pPattern++) << 4));
      j++;
    }
  return SANE_TRUE;
}

SANE_Bool
usb_high_cal_MD12i4o1_FillInDark (CalMD12 * cal, SANE_Word wMajor,
				  SANE_Word Minor, void * pDarkPattern)
{
  SANE_Byte * pPattern = (SANE_Byte *) pDarkPattern;
  SANE_Word j = 0;

  DBG (5,"usb_high_cal_MD12i4o1 FillInDark: wMajor=%d, Minor=%d\n",
	wMajor, Minor); 
  if (!cal->isPrepared)
    {
      DBG (5,"usb_high_cal_MD12i4o1::FillInDark !isPrepared");
      return SANE_FALSE;
    }
  if (cal->wDarkNeeded == 0)
    {
      DBG (5,"usb_high_cal_MD12i4o1::FillInDark wDarkNeeded==0");
      return SANE_FALSE;
    }

  while (j < cal->wWidth)
    {
      cal->pDarkLine[j++] += (double) (*(pPattern) & 0xf0);
      if (j >= cal->wWidth)
	break;
      cal->pDarkLine[j++] += (double) ((SANE_Byte) (*(pPattern++) << 4));
    }
  return SANE_TRUE;
}

void
usb_high_cal_MD12i8o8Mono_Calibrate (CalMD12 * cal, void * pSrc,
				    void * pTarget)
{
  SANE_Byte *pGraySrc = (SANE_Byte *) pSrc;
  SANE_Byte *pGrayTarget = (SANE_Byte *) pTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word i;

  DBG(5, "usb_high_cal_MDi8o8Mono_Calibrate Called\n");
  if (cal->pGammaTable == NULL)
    {
      SANE_Word wKWhiteLevel = cal->wKWhiteLevel >> 4;
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value = ((SANE_Word) (Base) * wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x00ff)
	    Value = 0x00ff;
	  pGrayTarget[i] = (SANE_Byte) (Value);
	}
    }
  else
    {
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value =
	    ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x0fff)
	    Value = 0x0fff;
	  pGrayTarget[i] = (SANE_Byte) (cal->pGammaTable[Value]);
	}
    }
}

void
usb_high_cal_MD12i8o8Rgb_Calibrate (CalMD12 * cal, void * pSrc,
				   void * pTarget)
{
  SANE_Byte *pGraySrc = (SANE_Byte *) pSrc;
  SANE_Byte *pRgbTarget = (SANE_Byte *) pTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word i;

  /*    DBG (5,"usb_high_cal_MDi8o8 Rgb Calibrator called \n"); */
  if (cal->pGammaTable == NULL)
    {
      SANE_Word wKWhiteLevel = cal->wKWhiteLevel >> 4;
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value = ((SANE_Word) (Base) * wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x00ff)
	    Value = 0x00ff;
	  *pRgbTarget = (SANE_Byte) (Value);
	  pRgbTarget += 3;
	}
    }
  else
    {
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value =
	    ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x0fff)
	    Value = 0x0fff;
	  *(pRgbTarget) = (SANE_Byte) (cal->pGammaTable[Value]);
	  pRgbTarget += 3;
	}
    }
}

void
usb_high_cal_MD12i4o1_Calibrate (CalMD12 * cal, void * pSrc,
				void * pTarget)
{
  SANE_Byte *p4BitSrc = (SANE_Byte *) pSrc;
  SANE_Byte *p1BitTarget = (SANE_Byte *) pTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word j = 0;
  int Count = 0;

  /*    DBG (5,"i4o1 calibrate called\n"); */
  *p1BitTarget = 0;
  while (j < cal->wWidth)
    {
      Base =
	(int) ((SANE_Word) (*p4BitSrc & 0xf0) << 4) - (int) (cal->pKDark[j]);
      if (Base < 0)
	Base = 0;
      Value = ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[j];
      if (Value > 0x0fff)
	Value = 0x0fff;
      if (Value >= cal->wThreshold)
	*(p1BitTarget) |= GrayMap[Count];
      Count++;
      j++;
      if (j >= cal->wWidth)
	break;
      Base =
	(int) ((SANE_Word) (*(p4BitSrc++) & 0x0f) << 8) -
	(int) (cal->pKDark[j]);
      if (Base < 0)
	Base = 0;
      Value = ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[j];
      if (Value > 0x0fff)
	Value = 0x0fff;
      if (Value >= cal->wThreshold)
	*(p1BitTarget) |= GrayMap[Count];
      Count++;
      if (Count >= 8)
	{
	  p1BitTarget++;
	  *p1BitTarget = 0;
	  Count = 0;
	}
      j++;
    }
}


/* scanner ------------------------------------------------------------ */


/* Enable this option for scan gray with green channel
   otherwise with three channels. */
#define GREENGRAY
#define CURMOTOR CAdaMotor1200
#define CURSENSOR300600 CAdaSensorCanon300600
#define CURSENSOR600	CAdaSensorCanon600

SANE_Bool
usb_high_scan_init (Mustek_Usb_Device * dev)
{
  dev->INI_MinExploseTime = A4CIS_MinExploseTime;
  dev->INI_300SkipsPerRow = A4CIS_SkipsPerRow;
  dev->INI_600SkipsPerRow = A4CIS_SkipsPerRow;
  dev->INI_JLines = A4CIS_JLines;
  dev->INI_KLines = A4CIS_KLines;
  dev->INI_KFilter = A4CIS_KFilter;
  dev->INI_KLoops = A4CIS_KLoops;
  dev->INI_PixelRateLines = 50;
  dev->INI_PixelRateFilts = 37;
  dev->INI_PowerDelayLines = A4CIS_PowerDelayLines;
  dev->INI_HomeLines = A4CIS_HomeLines;
  dev->INI_DarkLines = 50;
  dev->INI_KLevel = A4CIS_KLevel;
  dev->INI_MaxPowerDelay = A4CIS_MaxPowerDelay;
  dev->INI_MinPowerDelay = A4CIS_MaxPowerDelay;
  dev->INI_AdjustWay = 0;
  dev->INI_GreenBlackFactor = 0.0;
  dev->INI_BlueBlackFactor = 0.0;
  dev->INI_RedBlackFactor = 0.0;
  dev->INI_GrayBlackFactor = 0.0;
  dev->INI_GreenFactor = 0.0;
  dev->INI_BlueFactor = 0.0;
  dev->INI_RedFactor = 0.0;
  dev->INI_GrayFactor = 0.0;
  
  dev->INI_RedRgb600PGA = 0;
  dev->INI_GreenRgb600PGA = 0;
  dev->INI_BlueRgb600PGA = 0;
  dev->INI_Mono600PGA = 0;
  dev->INI_RedRgb300PGA = 0;
  dev->INI_GreenRgb300PGA = 0;
  dev->INI_BlueRgb300PGA = 0;
  dev->INI_Mono300PGA = 0;
  dev->INI_ExploseTime = 3200;
  dev->INI_RedRgb600PowerDelay = 0;
  dev->INI_GreenRgb600PowerDelay = 0;
  dev->INI_BlueRgb600PowerDelay = 0;
  dev->INI_RedMono600PowerDelay = 0;
  dev->INI_GreenMono600PowerDelay = 0;
  dev->INI_BlueMono600PowerDelay = 0;
  dev->INI_RedRgb300PowerDelay = 0;
  dev->INI_GreenRgb300PowerDelay = 0;
  dev->INI_BlueRgb300PowerDelay = 0;
  dev->INI_RedMono300PowerDelay = 0;
  dev->INI_GreenMono300PowerDelay = 0;
  dev->INI_BlueMono300PowerDelay = 0;
  dev->INI_Threshold = 128;
  
  dev->INI_TopRef = 0x10;
  dev->INI_FrontEnd = 0x10;
  dev->INI_RedOffset = 0x00;
  dev->INI_GreenOffset = 0x00;
  dev->INI_BlueOffset = 0x00;
  
  dev->INI_Rgb36BitBackTrack = 80;
  dev->INI_Rgb24BitBackTrack = 80;
  dev->INI_Mono8BitBackTrack = 80;
  dev->INI_Mono4BitBackTrack = 80;
  dev->INI_Mono1BitBackTrack = 80;

  dev->hinstOwner = 0;
  dev->isOpened = SANE_FALSE;
  dev->isPrepared = SANE_FALSE;
  dev->ExploseTime = 4000;
  dev->Width = 2550;
  dev->XDpi = 300;
  dev->YDpi = 300;
  dev->ScanMode = CRGB24ext;
  dev->BytesPerRow = 2550 * 3;
  dev->Dummy = 0;
  dev->BytesPerStrip = 2550;
  dev->pImageBuffer = NULL;
  dev->pRed = NULL;
  dev->pGreen = NULL;
  dev->pBlue = NULL;
  dev->pfGetLine = NULL;
  dev->pfBackTrack = NULL;
  dev->isAdjustedRgb600PowerDelay = SANE_FALSE;
  dev->isAdjustedMono600PowerDelay = SANE_FALSE;
  dev->isAdjustedRgb300PowerDelay = SANE_FALSE;
  dev->isAdjustedMono300PowerDelay = SANE_FALSE;
  dev->isEvaluePixelRate = SANE_FALSE;
  dev->RedRgb600PGA = 0;
  dev->GreenRgb600PGA = 0;
  dev->BlueRgb600PGA = 0;
  dev->Mono600PGA = 0;
  dev->RedRgb600PowerDelay = 0;
  dev->GreenRgb600PowerDelay = 0;
  dev->BlueRgb600PowerDelay = 0;
  dev->RedMono600PowerDelay = 0;
  dev->GreenMono600PowerDelay = 0;
  dev->BlueMono600PowerDelay = 0;
  dev->RedRgb300PGA = 0;
  dev->GreenRgb300PGA = 0;
  dev->BlueRgb300PGA = 0;
  dev->Mono300PGA = 0;
  dev->RedRgb300PowerDelay = 0;
  dev->GreenRgb300PowerDelay = 0;
  dev->BlueRgb300PowerDelay = 0;
  dev->RedMono300PowerDelay = 0;
  dev->GreenMono300PowerDelay = 0;
  dev->BlueMono300PowerDelay = 0;
  dev->PixelRate = 2000;
  dev->Threshold = 128;
  dev->pGammaTable = NULL;
  dev->SkipsPerRow = 0;

  dev->pRedCalibrator = NULL;
  dev->pGreenCalibrator = NULL;
  dev->pBlueCalibrator = NULL;
  dev->pMonoCalibrator = NULL;

  dev->bIs300600Switch = SANE_FALSE;
  dev->bAlreadyDetectCIS = SANE_FALSE;
  dev->chip = usb_low_init ();
  if (!dev->chip)
    {
      DBG (3, "usb_high_scan_init: usb_low_init returned NULL\n");
      return SANE_FALSE;
    }



  return SANE_TRUE;
}

void
usb_high_scan_exit (Mustek_Usb_Device * dev)
{
  usb_low_exit (dev->chip);
}

void
usb_high_scan_Init (Mustek_Usb_Device * dev, SANE_Int hinst,
		    SANE_String_Const pDeviceName)
{
  dev->hinstOwner = hinst;
  usb_high_scan_LoadPrivateProfile (dev, dev->hinstOwner);
  strcpy (dev->DeviceName, pDeviceName);
}

void
usb_high_scan_Release (Mustek_Usb_Device * dev)
{
  usb_high_scan_exit (dev);
}

#if 0
SCANNERSTATE usb_high_scan_GetScannerState (Mustek_Usb_Device * dev)
{
  if (dev->isOpened)
    {
      return ST_BUSY;
    }
  if (!usb_low_Open (dev->chip, dev->DeviceName, 0xa5))
    return ST_OFFLINE;
  usb_low_Close (dev->chip);
  return ST_READY;
}
#endif

void
usb_high_scan_Prepare (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_Prepare: dev=%p\n", dev);
  if (dev->isPrepared)
    {
      DBG (5,"usb_high_scan_Prepare if (isPrepared)");
      return;
    }
  if (dev->pImageBuffer)
    {
      free (dev->pImageBuffer);
    }
  dev->pImageBuffer = (SANE_Byte *) malloc (A4CIS_BytesPerStrip * 3);

  dev->pRed = dev->pImageBuffer;
  dev->pGreen = dev->pImageBuffer + A4CIS_BytesPerStrip;
  dev->pBlue = dev->pImageBuffer + A4CIS_BytesPerStrip * 2;

  dev->isPrepared = SANE_TRUE;
}

void
usb_high_scan_Clearup (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_Clearup: dev=%p\n", dev);
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_Clearup if (!isPrepared)");
      return;
    }
  if (dev->pImageBuffer)
    {
      free (dev->pImageBuffer);
    }
  dev->pImageBuffer = NULL;
  dev->pRed = NULL;
  dev->pGreen = NULL;
  dev->pBlue = NULL;

  dev->isPrepared = SANE_FALSE;
}

#if 0
void
usb_high_scan_GetInformation (Mustek_Usb_Device * dev,
			      PSCANNERINFORMATION pInfo,
			      PIOINFORMATION pIoInfo)
{
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_GetInformation isOpened");
      return;
    }
  if (!usb_low_Open (dev->chip, dev->DeviceName, 0xa5))
    {
      DBG (5,"usb_high_scan_GetInformation.. Open error");
      return;
    }
  usb_low_Close (dev->chip);

  pInfo->bMajor = 2;
  pInfo->bMinor = 0;
  strcpy (pInfo->sSensorName, "CANON 600 CIS");
  strcpy (pInfo->sAsicName, "MA1017");
  pInfo->wADCBits = 12;
  pInfo->wInputGammaBits = 12;
  pInfo->wOutputGammaBits = 8;
  pIoInfo->ioUSB.ioTag = IO_USB;
  pIoInfo->ioUSB.wNumOfEndPoints = 3;
  pIoInfo->ioUSB.uepUsbEndPoint[0] = UEP_BULKIN;
  pIoInfo->ioUSB.uepUsbEndPoint[1] = UEP_BULKOUT;
  pIoInfo->ioUSB.uepUsbEndPoint[2] = UEP_INTERRUPTIN;
}
#endif

void
usb_high_scan_TurnPower (Mustek_Usb_Device * dev, SANE_Bool isOn)
{
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_TurnPower isOpened");
      return;
    }
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
  if (isOn)
    {
      usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE);
      usb_low_TurnLampPower (dev->chip, SANE_TRUE);
    }
  else
    {
      usb_low_TurnLampPower (dev->chip, SANE_FALSE);
      /* usb_low_TurnPeripheralPower(SANE_FALSE); */
    }
  usb_low_Close (dev->chip);
}

#if 0
void
usb_high_scan_TurnLED (Mustek_Usb_Device * dev, SANE_Bool isOn)
{
  DBG (5, "usb_high_scan_TurnLED: dev=%p, isOn=%d\n", dev, isOn);
  if (dev->isOpened) /* ??? */
    {
      DBG (5,"usb_high_scan_TurnLED isOpened");
      return;
    }
  DBG (5,"usb_high_scan_TurnLED: not supported");
}
#endif

void
usb_high_scan_BackHome (Mustek_Usb_Device * dev)
{
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_BackHome isOpened");
      return;
    }
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
  usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime);
  usb_mid_motor1200_PrepareHome (dev->chip);
  usb_low_Close (dev->chip);
}

#if 0
SENSORSTATE usb_high_scan_IsCoverOpen (Mustek_Usb_Device * dev,
				       SANE_Bool Reserve)
{
  SENSORSTATE csCover;

  DBG (5,"usb_high_scan_IsCoverOpen: dev=%p, Reserve=%d\n",
       dev, Reserve);
 
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_IsCoverOpen isOpened");
      return SS_UNKNOW;
    }
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
  if (usb_low_GetSTOP (dev->chip))
    {
      csCover = SS_ON;
    }
  else
    {
      csCover = SS_OFF;
    }
  usb_low_Close (dev->chip);

  return csCover;
}
#endif

#if 0
SENSORSTATE usb_high_scan_IsCarriageHome (Mustek_Usb_Device * dev, 
					  SANE_Bool Reserve)
{
  SENSORSTATE csCover;

  DBG (5,"usb_high_scan_IsCarriageHome: dev=%p, Reserve=%d\n",
       dev, Reserve);

  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_IsCarriageHome isOpened");
      return SS_UNKNOW;
    }
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
  if (!usb_low_GetHomeSensor (dev->chip))
    {
      csCover = SS_OFF;
    }
  else
    {
      csCover = SS_ON;
    }
  usb_low_Close (dev->chip);

  return csCover;
}
#endif

void
usb_high_scan_SetThreshold (Mustek_Usb_Device * dev, SANE_Byte threshold)
{
  DBG (5,"usb_high_scan_SetThreshold: dev=%p, threshold=%d\n",
       dev, threshold);

  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_SetThreshold isOpened");
      return;
    }

  dev->Threshold = threshold;
}

void
usb_high_scan_EmbedGamma (Mustek_Usb_Device * dev, SANE_Word * pGammaTable)
{
  DBG (5,"usb_high_scan_EmbedGamma: dev=%p, pGammaTable=%d\n",
       dev, pGammaTable);
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_EmbedGamma isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_EmbedGamma !isPrepared");
      return;
    }

  dev->pGammaTable = pGammaTable;
}

void
usb_high_scan_Reset (Mustek_Usb_Device * dev)
{
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_Reset isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_Reset !isPrepared");
      return;
    }
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);

  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_high_scan_InitASIC (dev, SR_CANON300600);
  else
    usb_high_scan_InitASIC (dev, SR_CANON600);
  /* */
  usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime);
  usb_mid_motor1200_PrepareHome (dev->chip);
  usb_low_Close (dev->chip);

  usb_high_scan_SetThreshold (dev, dev->INI_Threshold);
  usb_high_scan_EmbedGamma (dev, NULL);
  dev->isAdjustedRgb600PowerDelay = SANE_FALSE;
  dev->isAdjustedMono600PowerDelay = SANE_FALSE;
  dev->isAdjustedRgb300PowerDelay = SANE_FALSE;
  dev->isAdjustedMono300PowerDelay = SANE_FALSE;
  dev->isEvaluePixelRate = SANE_FALSE;
}

void
usb_high_scan_WaitCarriageHome (Mustek_Usb_Device * dev)
{
  if (!usb_low_GetHomeSensor (dev->chip))
    {
      usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime);
      usb_mid_motor1200_PrepareHome (dev->chip);
      while (!usb_low_GetHomeSensor (dev->chip))
	{
	  usleep (18 * 1000);
	}
    }
  usb_low_MoveMotorHome (dev->chip, SANE_FALSE, SANE_FALSE);
  /* No Motor & Forward */
}

void
usb_high_scan_HardwareCalibration (Mustek_Usb_Device * dev)
{
  usb_high_scan_SafeForward (dev, dev->INI_HomeLines);
  switch (dev->INI_AdjustWay)
    {
    case 1:
      switch (dev->ScanMode)
	{
	case CRGB48:
	case CRGB36ext:
	case CRGB24ext:
	  /* Modified by Eden */
	  if (dev->bIs300600Switch)
	    {
	      if (usb_mid_c300600_is600Mode
		  (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
		  dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
		  dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
		  usb_high_scan_AdjustRgb600PowerDelay (dev);
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
		  dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
		  dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
		  usb_high_scan_AdjustRgb300PowerDelay (dev);
		}
	    }
	  else
	    {
	      if (usb_mid_c600_is600Mode
		  (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
		  dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
		  dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
		  usb_high_scan_AdjustRgb600PowerDelay (dev);
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
		  dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
		  dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
		  usb_high_scan_AdjustRgb300PowerDelay (dev);
		}
	    }
	  /* */
	  break;
	case CGRAY8ext:
	case CTEXText:
	case CTEXT:
	  /* Modified by Eden */
	  if (dev->bIs300600Switch)
	    {
	      if (usb_mid_c300600_is600Mode
		  (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono600PGA = dev->INI_Mono600PGA;
		  usb_high_scan_EvaluePixelRate (dev);
		  usb_high_scan_AdjustMono600PowerDelay (dev);
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono300PGA = dev->INI_Mono300PGA;
		  usb_high_scan_EvaluePixelRate (dev);
		  usb_high_scan_AdjustMono300PowerDelay (dev);
		}
	    }
	  else
	    {
	      if (usb_mid_c600_is600Mode
		  (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono600PGA = dev->INI_Mono600PGA;
		  usb_high_scan_EvaluePixelRate (dev);
		  usb_high_scan_AdjustMono600PowerDelay (dev);
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono300PGA = dev->INI_Mono300PGA;
		  usb_high_scan_EvaluePixelRate (dev);
		  usb_high_scan_AdjustMono300PowerDelay (dev);
		}
	    }
	  /* */
	  break;
	default:
	  break;
	}
      break;
    default:
      dev->ExploseTime = dev->INI_ExploseTime;
      dev->RedRgb600PowerDelay = dev->INI_RedRgb600PowerDelay;
      dev->GreenRgb600PowerDelay = dev->INI_GreenRgb600PowerDelay;
      dev->BlueRgb600PowerDelay = dev->INI_BlueRgb600PowerDelay;
      dev->RedMono600PowerDelay = dev->INI_RedMono600PowerDelay;
      dev->GreenMono600PowerDelay = dev->INI_GreenMono600PowerDelay;
      dev->BlueMono600PowerDelay = dev->INI_BlueMono600PowerDelay;
      dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
      dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
      dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
      dev->Mono600PGA = dev->INI_Mono600PGA;
      dev->RedRgb300PowerDelay = dev->INI_RedRgb300PowerDelay;
      dev->GreenRgb300PowerDelay = dev->INI_GreenRgb300PowerDelay;
      dev->BlueRgb300PowerDelay = dev->INI_BlueRgb300PowerDelay;
      dev->RedMono300PowerDelay = dev->INI_RedMono300PowerDelay;
      dev->GreenMono300PowerDelay = dev->INI_GreenMono300PowerDelay;
      dev->BlueMono300PowerDelay = dev->INI_BlueMono300PowerDelay;
      dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
      dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
      dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
      dev->Mono300PGA = dev->INI_Mono300PGA;
      break;
    }
}

void
usb_high_scan_LineCalibration (Mustek_Usb_Device * dev)
{
  switch (dev->ScanMode)
    {
    case CRGB48:
      usb_high_scan_PrepareRgb48Bit (dev);
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationRgb48Bit (dev);
      break;
    case CRGB36ext:
      usb_high_scan_PrepareRgb36Bit (dev);
      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationRgb36Bit (dev);
      break;
    case CRGB24ext:
      usb_high_scan_PrepareRgb24Bit (dev);
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationRgb24Bit (dev);
      break;
    case CGRAY8ext:
      usb_high_scan_PrepareMono8Bit (dev);
      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationMono8Bit (dev);
      break;
    case CTEXText:
      usb_high_scan_PrepareMono4Bit (dev);
      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationMono4Bit (dev);
      break;
    case CTEXT:
      usb_high_scan_PrepareMono1Bit (dev);
      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */
      usb_high_scan_CalibrationMono1Bit (dev);
      break;
    default:
      break;
    }
}

void
usb_high_scan_PrepareScan (Mustek_Usb_Device * dev)
{
  switch (dev->ScanMode)
    {
    case CRGB48:
      usb_high_scan_PrepareRgb48Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetRgb48BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackRgb48Bit;

      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);

	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareRgbY (dev->chip, dev->YDpi);
      break;
    case CRGB36ext:
      usb_high_scan_PrepareRgb36Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetRgb36BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackRgb36Bit;

      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareRgbY (dev->chip, dev->YDpi);
      break;
    case CRGB24ext:
      usb_high_scan_PrepareRgb24Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetRgb24BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackRgb24Bit;

      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c300600_PrepareRgbX (dev->chip,
					      dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareRgbSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareRgbSignal300Dpi (dev);
	  usb_mid_c600_PrepareRgbX (dev->chip,
					   dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareRgbY (dev->chip, dev->YDpi);
      break;
    case CGRAY8ext:
      usb_high_scan_PrepareMono8Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetMono8BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackMono8Bit;
      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareMonoY (dev->chip, dev->YDpi);
      break;
    case CTEXText:
      usb_high_scan_PrepareMono4Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetMono4BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackMono4Bit;

      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareMonoY (dev->chip, dev->YDpi);
      break;
    case CTEXT:
      usb_high_scan_PrepareMono1Bit (dev);
      dev->pfGetLine = &usb_high_scan_GetMono1BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackMono1Bit;

      /* Modified by Eden */
      if (dev->bIs300600Switch)
	{
	  if (usb_mid_c300600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c300600_PrepareMonoX (dev->chip,
					       dev->XDpi);
	}
      else
	{
	  if (usb_mid_c600_is600Mode
	      (dev->chip, dev->XDpi))
	    usb_high_scan_PrepareMonoSignal600Dpi (dev);
	  else
	    usb_high_scan_PrepareMonoSignal300Dpi (dev);
	  usb_mid_c600_PrepareMonoX (dev->chip,
					    dev->XDpi);
	}
      /* */

      usb_mid_motor1200_PrepareMonoY (dev->chip, dev->YDpi);
      break;
    default:
      break;
    }
}

void
usb_high_scan_ScanSuggest (Mustek_Usb_Device * dev, PTARGETIMAGE pTarget,
			   PSUGGESTSETTING pSuggest)
{
  /* Looking up Optical Y Resolution */
  int i;
  for (i = 0; usb_mid_motor1200_s_wOpticalYDpi[i] != 0; i++)
    {
      if (usb_mid_motor1200_s_wOpticalYDpi[i] <= pTarget->wDpi)
	{
	  pSuggest->wYDpi = usb_mid_motor1200_s_wOpticalYDpi[i];
	  break;
	}
    }
  if (usb_mid_motor1200_s_wOpticalYDpi[i] == 0)
    {
      i--;
      pSuggest->wYDpi = usb_mid_motor1200_s_wOpticalYDpi[i];
    }

  /* Looking up Optical X Resolution */
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    {
      for (i = 0; usb_mid_c300600_s_wOpticalXDpi[i] != 0; i++)
	{
	  if (usb_mid_c300600_s_wOpticalXDpi[i] <= pTarget->wDpi)
	    {
	      pSuggest->wXDpi = usb_mid_c300600_s_wOpticalXDpi[i];
	      break;
	    }
	}
      if (usb_mid_c300600_s_wOpticalXDpi[i] == 0)
	{
	  i--;
	  pSuggest->wXDpi = usb_mid_c300600_s_wOpticalXDpi[i];
	}
    }
  else
    {
      for (i = 0; usb_mid_c600_s_wOpticalXDpi[i] != 0; i++)
	{
	  if (usb_mid_c600_s_wOpticalXDpi[i] <= pTarget->wDpi)
	    {
	      pSuggest->wXDpi = usb_mid_c600_s_wOpticalXDpi[i];
	      break;
	    }
	}
      if (usb_mid_c600_s_wOpticalXDpi[i] == 0)
	{
	  i--;
	  pSuggest->wXDpi = usb_mid_c600_s_wOpticalXDpi[i];
	}
    }
  /* */
  pSuggest->wX = (SANE_Word) (((SANE_Word) (pTarget->wX) 
			       * (SANE_Word) (pSuggest->wXDpi)) 
			      / (SANE_Word) (pTarget->wDpi));
  pSuggest->wY = (SANE_Word) (((SANE_Word) (pTarget->wY) 
			       * (SANE_Word) (pSuggest->wYDpi)) 
			      / (SANE_Word) (pTarget->wDpi));
  pSuggest->wWidth = (SANE_Word) (((SANE_Word) (pTarget->wWidth) 
				   * (SANE_Word) (pSuggest->wXDpi)) 
				  / (SANE_Word) (pTarget->wDpi));
  pSuggest->wHeight = (SANE_Word) (((SANE_Word) (pTarget->wHeight) 
				    * (SANE_Word) (pSuggest->wYDpi)) 
				   / (SANE_Word) (pTarget->wDpi));

  if (pTarget->isOptimalSpeed)
    {
      switch (pTarget->cmColorMode)
	{
	case CRGB48:
	  pSuggest->cmScanMode = CRGB48;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 6;
	  break;
	case CRGB36:
	  pSuggest->cmScanMode = CRGB36ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 6;
	  break;
	case CRGB24:
	  pSuggest->cmScanMode = CRGB24ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 3;
	  break;
	case CGRAY8:
	  pSuggest->cmScanMode = CGRAY8ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth);
	  break;
	case CTEXT:
	  pSuggest->cmScanMode = CTEXText;
	  pSuggest->wWidth /= 8;
	  pSuggest->wWidth *= 8;
	  pSuggest->dwBytesPerRow = ((SANE_Word) (pSuggest->wWidth) + 7) / 8;
	  break;
	default:
	  break;
	}
    }
  else
    {
      switch (pTarget->cmColorMode)
	{
	case CRGB48:
	  pSuggest->cmScanMode = CRGB48;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 6;
	  break;
	case CRGB36:
	  pSuggest->cmScanMode = CRGB36ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 6;
	  break;
	case CRGB24:
	  pSuggest->cmScanMode = CRGB24ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 3;
	  break;
	case CGRAY8:
	  pSuggest->cmScanMode = CGRAY8ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth);
	  break;
	case CTEXT:
	  pSuggest->cmScanMode = CTEXText;
	  pSuggest->wWidth /= 8;
	  pSuggest->wWidth *= 8;
	  pSuggest->dwBytesPerRow = ((SANE_Word) (pSuggest->wWidth) + 7) / 8;
	  break;
	default:
	  break;
	}
    }
}

void
usb_high_scan_SetupScan (Mustek_Usb_Device * dev, COLORMODE ColorMode,
			 SANE_Word XDpi, SANE_Word YDpi, SANE_Bool isInvert,
			 SANE_Word X, SANE_Word Y, SANE_Word Width)
{
  /* Modified by Eden */
  SANE_Byte *pBuffer;
  static SANE_Word L_Temp = 0, R_Temp = 0;
  int i;
  SANE_Word UpperBound;
  SANE_Word LeftBound;

  DBG (5,"usb_high_scan_SetupScan: isInvert=%d\n", isInvert);
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_SetupScan isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_SetupScan !isPrepared");
      return;
    }

  if (!dev->bAlreadyDetectCIS)
    {
      pBuffer = NULL;
      L_Temp = 0;
      R_Temp = 0;

      DBG (5,"usb_high_scan_SetupScan Detecting CIS\n");
      pBuffer = (SANE_Byte *) malloc (A4CIS_BytesPerStrip);
      for (i = 0; i < 5400; i++)
	pBuffer[i] = 0xaa;

      dev->ScanMode = CGRAY8ext;
      dev->XDpi = 600;
      dev->YDpi = 1200;
      dev->Width = 5400;
      usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
      dev->isOpened = SANE_TRUE;
      usb_high_scan_InitASIC (dev, SR_CANON600);
      usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE);
      usb_low_EnableMotor (dev->chip, SANE_TRUE);	/* Enable Motor */
      usb_low_TurnLampPower (dev->chip, SANE_TRUE);
      usb_low_InvertImage (dev->chip, SANE_FALSE);
      usb_low_SetImageDpi (dev->chip, SANE_TRUE, SW_P6P6);
      dev->BytesPerStrip = A4CIS_600AdjustLength;
      dev->BytesPerRow = 5400;
      dev->Dummy = 0;

      usb_high_scan_WaitCarriageHome (dev);
      usb_high_scan_HardwareCalibration (dev);
      usb_high_scan_PrepareScan (dev);

      /* Get Data */
      usb_low_StartRowing (dev->chip);
      DBG (5,"usb_low_StartRowing done\n");
      usb_low_GetRow (dev->chip, pBuffer);
      DBG (5,"usb_low_GetRow done.\n");
      usb_low_StopRowing (dev->chip);
      DBG (5,"usb_low_StopRowing done.\n");
      /* Calculate */
      for (i = 0; i < 256; i++)
	L_Temp = L_Temp + pBuffer[512 + i];
      for (i = 0; i < 256; i++)
	R_Temp = R_Temp + pBuffer[3500 + i];

      L_Temp = L_Temp / 256;
      R_Temp = R_Temp / 256;

      /* Release Resource */
      usb_low_Close (dev->chip);
      free (pBuffer);
      pBuffer = NULL;
      DBG (5,"usb_high_scan_SetupScan Detect Done.\n");
    }
  /* 300/600 switch CIS or 600 CIS */
  if (R_Temp > 50)
    dev->bIs300600Switch = SANE_FALSE;
  else
    dev->bIs300600Switch = SANE_TRUE;

  /*    dev->bIs300600Switch = SANE_FALSE; */
  usb_low_SetCIS (dev->chip, dev->bIs300600Switch);
  dev->bAlreadyDetectCIS = SANE_TRUE;
  /* */
  usb_low_Open (dev->chip, dev->DeviceName, 0xa5);
  dev->isOpened = SANE_TRUE;

  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_high_scan_InitASIC (dev, SR_CANON300600);
  else
    usb_high_scan_InitASIC (dev, SR_CANON600);
  /* */

  usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE);
  usb_low_EnableMotor (dev->chip, SANE_TRUE);	/* Enable Motor */
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);
  usb_low_InvertImage (dev->chip, SANE_FALSE);

  dev->ScanMode = ColorMode;
  dev->XDpi = XDpi;
  dev->YDpi = YDpi;
  dev->Width = Width;

  /* Compute necessary variable */
  UpperBound = (SANE_Word) (((SANE_Word) (Y) * 600L) / (SANE_Word) (dev->YDpi)) + dev->INI_JLines;	/* in 600dpi */

  switch (dev->ScanMode)
    {
    case CRGB48:
      dev->BytesPerRow = 6 * dev->Width;
      break;
    case CRGB36ext:
      dev->BytesPerRow = 6 * dev->Width;
      break;
    case CRGB24ext:
      dev->BytesPerRow = 3 * dev->Width;
      break;
    case CGRAY8ext:
      dev->BytesPerRow = dev->Width;
      break;
    case CTEXText:
      dev->BytesPerRow = (dev->Width + 7) / 8;
      break;
    case CTEXT:
      dev->BytesPerRow = (dev->Width + 7) / 8;
      break;
    default:
      break;
    }

  /* Modified by Eden */
  if (dev->bIs300600Switch)
    {
      if (usb_mid_c300600_is600Mode
	  (dev->chip, dev->XDpi))
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 600L) / (SANE_Word) (dev->XDpi)) + dev->INI_600SkipsPerRow;	/* in 600dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (
		    ((SANE_Word) (LeftBound % 32) * (SANE_Word) (dev->XDpi) +
		     300L) / 600L);
	}
      else
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 300L) / (SANE_Word) (dev->XDpi)) + dev->INI_300SkipsPerRow;	/* in 300dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (
		    ((SANE_Word) (LeftBound % 32) * (SANE_Word) (dev->XDpi) +
		     150L) / 300L);
	}
    }
  else
    {
      if (usb_mid_c600_is600Mode
	  (dev->chip, dev->XDpi))
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 600L) / (SANE_Word) (dev->XDpi)) + dev->INI_600SkipsPerRow;	/* in 600dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (
		    ((SANE_Word) (LeftBound % 32) * (SANE_Word) (dev->XDpi) +
		     300L) / 600L);
	}
      else
	{
	  LeftBound = (SANE_Word) (((SANE_Word) X) * 300L / (SANE_Word) dev->XDpi) + dev->INI_300SkipsPerRow;	/* in 300dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (
		    ((SANE_Word) LeftBound % 32 * ((SANE_Word) dev->XDpi) +
		     150L) / 300L);
	}
    }
  /* */
  dev->Dummy = (LeftBound / 32) * 32;

  switch (dev->ScanMode)
    {
    case CRGB48:
      dev->SkipsPerRow = ((dev->SkipsPerRow + 1) / 2) * 3;
      dev->BytesPerStrip =
	dev->SkipsPerRow + (dev->Width * 3 + 1) / 2;
      break;
    case CRGB36ext:
      dev->SkipsPerRow = ((dev->SkipsPerRow + 1) / 2) * 3;
      dev->BytesPerStrip =
	dev->SkipsPerRow + (dev->Width * 3 + 1) / 2;
      break;
    case CRGB24ext:
      dev->BytesPerStrip = dev->SkipsPerRow + dev->Width;
      break;
    case CGRAY8ext:
      dev->BytesPerStrip = dev->SkipsPerRow + dev->Width;
      break;
    case CTEXText:
      dev->SkipsPerRow = (dev->SkipsPerRow + 1) / 2;
      dev->BytesPerStrip =
	dev->SkipsPerRow + (dev->Width + 1) / 2;
      break;
    case CTEXT:
      dev->SkipsPerRow = (dev->SkipsPerRow + 7) / 8;
      dev->BytesPerStrip =
	dev->SkipsPerRow + (dev->Width + 7) / 8;
      break;
    default:
      break;
    }
  dev->BytesPerStrip = ((dev->BytesPerStrip + 1) / 2) * 2;	/* make BytesPerStrip is as 2n to advoid 64n+1 */

  usb_high_scan_WaitCarriageHome (dev);
  usb_high_scan_HardwareCalibration (dev);
  usb_high_scan_LineCalibration (dev);
  usb_high_scan_StepForward (dev, UpperBound);
  usb_high_scan_PrepareScan (dev);
  usb_low_StartRowing (dev->chip);
}

void
usb_high_scan_GetRows (Mustek_Usb_Device * dev, SANE_Byte * lpBlock, SANE_Word Rows,
		       SANE_Bool isOrderInvert)
{
  if (!dev->isOpened)
    {
      DBG (5,"usb_high_scan_GetRows !isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_GetRows !isPrepared");
      return;
    }
  while (Rows > 0)
    {
      (*dev->pfGetLine) (dev, lpBlock, isOrderInvert);
      lpBlock += dev->BytesPerRow;
      Rows--;
    }
}

#if 0
void
usb_high_scan_BackTrack (Mustek_Usb_Device * dev)
{
  if (!dev->isOpened)
    {
      DBG (5,"usb_high_scan_BackTrack !isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_BackTrack !isPrepared");
      return;
    }
  if (!dev->pfBackTrack)
    {
      DBG (5,"usb_high_scan_BackTrack ... no callback func\n");
      return;
    }
  (*dev->pfBackTrack) (dev);
}
#endif

void
usb_high_scan_StopScan (Mustek_Usb_Device * dev)
{
  if (!dev->isOpened)
    {
      DBG (5,"usb_high_scan_StopScan !isOpened");
      return;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_StopScan !isPrepared");
    }
  switch (dev->ScanMode)
    {
    case CRGB48:
    case CRGB36ext:
    case CRGB24ext:
      usb_high_cal_MD12_Clearup (dev->pBlueCalibrator);
      usb_high_cal_MD12_Release (dev->pBlueCalibrator);
      free (dev->pBlueCalibrator);
      dev->pBlueCalibrator = NULL;
      usb_high_cal_MD12_Clearup (dev->pGreenCalibrator);
      usb_high_cal_MD12_Release (dev->pGreenCalibrator);
      free (dev->pGreenCalibrator);
      dev->pGreenCalibrator = NULL;
      usb_high_cal_MD12_Clearup (dev->pRedCalibrator);
      usb_high_cal_MD12_Release (dev->pRedCalibrator);
      free (dev->pRedCalibrator);
      dev->pRedCalibrator = NULL;
      break;
    case CGRAY8ext:
    case CTEXText:
      usb_high_cal_MD12_Clearup (dev->pMonoCalibrator);
      usb_high_cal_MD12_Release (dev->pMonoCalibrator);
      free (dev->pMonoCalibrator);
      dev->pMonoCalibrator = NULL;
      break;
    case CTEXT:
      break;
    default:
      break;
    }

  usb_low_StopRowing (dev->chip);
  usb_low_TurnLampPower (dev->chip, SANE_FALSE);
  usb_low_Close (dev->chip);
  dev->isOpened = SANE_FALSE;
}

void
usb_high_scan_LoadPrivateProfile (Mustek_Usb_Device * dev, SANE_Int hinst)
{
  DBG (5, "usb_high_scan_LoadPrivateProfile: dev=%p, hinst=%d\n",
       dev, hinst);
  dev->INI_MinExploseTime = 4992;
  dev->INI_600SkipsPerRow = 72;
  dev->INI_300SkipsPerRow = 56;
  dev->INI_JLines = 154;
  dev->INI_KLines = 16;
  dev->INI_KFilter = 8;
  dev->INI_KLoops = 2;
  dev->INI_PixelRateLines = 50;
  dev->INI_PixelRateFilts = 37;
  dev->INI_PowerDelayLines = 2;
  dev->INI_HomeLines = 160;
  dev->INI_DarkLines = 50;
  dev->INI_KLevel = 245;
  dev->INI_MaxPowerDelay = 240;
  dev->INI_MinPowerDelay = 136;
  dev->INI_AdjustWay = 1;
  dev->INI_GreenBlackFactor = -2.0;
			    /*-1.48;*/
  dev->INI_BlueBlackFactor = 0.0;
			  /*-1.69;*/
  dev->INI_RedBlackFactor = 0.0;	/*0.7; */
  dev->INI_GrayBlackFactor = 0.0;
			  /*-1.84;*/

  /* For TestAP... */
  /*dev->INI_GreenFactor=0.7556832608;
     dev->INI_BlueFactor=0.7527943848;
     dev->INI_RedFactor=0.754877035;
     dev->INI_GrayFactor=0.756737835; */
  /* For Release... */
  dev->INI_GreenFactor = 0.82004;
  dev->INI_BlueFactor = 0.84954;
  dev->INI_RedFactor = 0.826375;
  dev->INI_GrayFactor = 0.833375;


  dev->INI_RedRgb600PGA = 8;
  dev->INI_GreenRgb600PGA = 8;
  dev->INI_BlueRgb600PGA = 8;
  dev->INI_Mono600PGA = 8;
  dev->INI_RedRgb300PGA = 8;
  dev->INI_GreenRgb300PGA = 8;
  dev->INI_BlueRgb300PGA = 8;
  dev->INI_Mono300PGA = 8;

  dev->INI_ExploseTime = 9024;
  dev->INI_RedRgb600PowerDelay = 80;
  dev->INI_GreenRgb600PowerDelay = 80;
  dev->INI_BlueRgb600PowerDelay = 80;
  dev->INI_RedMono600PowerDelay = 80;
  dev->INI_GreenMono600PowerDelay = 80;
  dev->INI_BlueMono600PowerDelay = 80;
  dev->INI_RedRgb300PowerDelay = 80;
  dev->INI_GreenRgb300PowerDelay = 80;
  dev->INI_BlueRgb300PowerDelay = 80;
  dev->INI_RedMono300PowerDelay = 80;
  dev->INI_GreenMono300PowerDelay = 80;
  dev->INI_BlueMono300PowerDelay = 80;
  dev->INI_Threshold = 128;

  dev->INI_TopRef = 128;
  dev->INI_FrontEnd = 16;
  dev->INI_RedOffset = 0;
  dev->INI_GreenOffset = 0;
  dev->INI_BlueOffset = 0;

  dev->INI_Rgb36BitBackTrack = 80;
  dev->INI_Rgb24BitBackTrack = 80;
  dev->INI_Mono8BitBackTrack = 80;
  dev->INI_Mono4BitBackTrack = 80;
  dev->INI_Mono1BitBackTrack = 80;
}

void
usb_high_scan_InitASIC (Mustek_Usb_Device * dev, SENSOR Sensor)
{
  /*  char Dir[140];
  char Section[40];
  char Byte[10];*/
  SANE_Byte CcdDpi = 0;
  SANE_Byte Select = 0;
  SANE_Byte Adjust = 0;
  SANE_Byte Pin = 0;
  SANE_Byte Motor = 0;
  SANE_Bool FixPattern = SANE_FALSE;
  SANE_Byte ADTiming = 0;

  DBG (5, "usb_high_scan_InitASIC: \n");
  switch (Sensor)
    {
    case SR_TOSHIBA600:
      CcdDpi = 32;
      Select = 240;
      Adjust = 0;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 0;
      break;
    case SR_CANON300:
      CcdDpi = 232;
      Select = 232;
      Adjust = 0;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      break;
      /* Modified by Eden */
    case SR_CANON300600:
      CcdDpi = 232;
      Select = 232;
      Adjust = 64;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      break;
    case SR_CANON600:
      CcdDpi = 232;
      Select = 232;
      Adjust = 64;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      break;
    default:
      break;
    }
  usb_low_AdjustTiming (dev->chip, Adjust);
  usb_low_SelectTiming (dev->chip, Select);
  usb_low_SetTiming (dev->chip, CcdDpi);
  usb_low_SetSRAMBank (dev->chip, BS_16K);
  usb_low_SetASICIoPins (dev->chip, Pin);
  usb_low_SetRGBSelPins (dev->chip, Pin);
  usb_low_SetMotorSignal (dev->chip, Motor);
  usb_low_SetTestSRAMMode (dev->chip, SANE_FALSE);
  usb_low_SetFixPattern (dev->chip, FixPattern);
  usb_low_SetADTiming (dev->chip, ADTiming);
}

SANE_Byte
usb_high_scan_EvalueMaxLevel (Mustek_Usb_Device * dev, SANE_Word SampleLines,
			  int SampleLength)
{
  SANE_Byte MaxLevel = 0;
  SANE_Word i;
  int j;
  SampleLength -= 20;
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < SampleLines; i++)
    {
      usb_low_GetRow (dev->chip, dev->pGreen);
      for (j = 20; j < SampleLength; j++)
	{
	  if (MaxLevel < dev->pGreen[j])
	    MaxLevel = dev->pGreen[j];
	}
    }
  usb_low_StopRowing (dev->chip);
  return MaxLevel;
}

/* Binary Search for Single Channel Power Delay */
void
usb_high_scan_BSSCPowerDelay (Mustek_Usb_Device * dev,
			  usb_high_PPOWERDELAYFUNC pSetPowerDelay,
			  SIGNALSTATE * ssCurrent, SANE_Byte * Target, SANE_Byte Max,
			  SANE_Byte Min, SANE_Byte Threshold, int Length)
{
  SANE_Byte MaxLevel;
  SANE_Byte MaxMax = Max;
  SANE_Byte MinMin = Min;
  *Target = (Max + Min) / 2;
  (*pSetPowerDelay) (dev->chip, *Target);
  while (*Target != Min)
    {
      MaxLevel =
	usb_high_scan_EvalueMaxLevel (dev, dev->INI_PowerDelayLines, Length);
      if (MaxLevel > Threshold)
	{
	  Min = *Target;
	  *Target = (Max + Min) / 2;
	  *ssCurrent = SG_BRIGHTER;
	}
      else if (MaxLevel < Threshold)
	{
	  Max = *Target;
	  *Target = (Max + Min) / 2;
	  *ssCurrent = SG_DARKER;
	}
      else if (MaxLevel == Threshold)
	{			/* Found. */
	  *ssCurrent = SG_EQUAL;
	  return;
	}
      (*pSetPowerDelay) (dev->chip, *Target);
    }
  /* Fail... */
  if (Max == MaxMax || Min == MinMin)
    {				/* Boundary check */
      if (Max == MaxMax)	/*Target on max side */
	*Target = MaxMax;
      else
	*Target = MinMin;
      (*pSetPowerDelay) (dev->chip, *Target);
      MaxLevel =
	usb_high_scan_EvalueMaxLevel (dev, dev->INI_PowerDelayLines, Length);
      if (MaxLevel > Threshold)
	{
	  *ssCurrent = SG_BRIGHTER;
	}
      else if (MaxLevel < Threshold)
	{
	  *ssCurrent = SG_DARKER;
	}
      else if (MaxLevel == Threshold)
	{
	  *ssCurrent = SG_EQUAL;
	}
    }
  else
    {				/* Fail... will always on mimnum side, make it darker */
      Target++;
      *ssCurrent = SG_DARKER;
    }
}

#if 0
static SANE_Bool LSACPDMap[3][3] =
{
  {
  SANE_FALSE, SANE_FALSE, SANE_FALSE}
  ,
  {
  SANE_FALSE, SANE_FALSE, SANE_TRUE}
  ,
  {
  SANE_FALSE, SANE_TRUE, SANE_FALSE}
};
#endif

/* Linear Search for All Channel Power Delay */
#if 0
void
usb_high_scan_LSACPowerDelay (Mustek_Usb_Device * dev, SANE_Byte * RedTarget,
			  SANE_Byte * GreenTarget, SANE_Byte * BlueTarget,
			  SANE_Byte Threshold, int Length)
{
  double ExploseTime = (double) (dev->ExploseTime) / 64.0;
  double RedLightUp = ExploseTime - (double) (*RedTarget);
  double GreenLightUp = ExploseTime - (double) (*GreenTarget);
  double BlueLightUp = ExploseTime - (double) (*BlueTarget);
  double MiniLightUp = min (RedLightUp, min (GreenLightUp, BlueLightUp));
  double RedStep, GreenStep, BlueStep, RedPowerDelay, GreenPowerDelay,
    BluePowerDelay;
  SIGNALSTATE ssCurrent = SG_UNKNOW;
  SIGNALSTATE ssLast = SG_UNKNOW;
  SANE_Byte MaxLevel;

  if (MiniLightUp <= 0.0)
    return;
  RedStep = (RedLightUp / MiniLightUp);
  GreenStep = (GreenLightUp / MiniLightUp);
  BlueStep = (BlueLightUp / MiniLightUp);
  RedPowerDelay = ExploseTime - RedLightUp / 3.0;
  GreenPowerDelay = ExploseTime - GreenLightUp / 3.0;
  BluePowerDelay = ExploseTime - BlueLightUp / 3.0;

  /* Adjust AllPD */
  while (SANE_TRUE)
    {
      usb_low_SetRedPD (dev->chip, (SANE_Byte) (RedPowerDelay));
      usb_low_SetGreenPD (dev->chip, (SANE_Byte) (GreenPowerDelay));
      usb_low_SetBluePD (dev->chip, (SANE_Byte) (BluePowerDelay));
      MaxLevel =
	usb_high_scan_EvalueMaxLevel (dev, dev->INI_PowerDelayLines, Length);
      ssLast = ssCurrent;
      if (MaxLevel == Threshold)
	{
	  ssCurrent = SG_EQUAL;
	  break;
	}
      ssCurrent = (MaxLevel > Threshold) ? SG_BRIGHTER : SG_DARKER;
      if (LSACPDMap[ssCurrent][ssLast])
	{			/* Found */
	  break;
	}
      if (ssCurrent == SG_BRIGHTER)
	{
	  RedPowerDelay += RedStep;
	  GreenPowerDelay += GreenStep;
	  BluePowerDelay += BlueStep;
	  if (BluePowerDelay > ExploseTime || GreenPowerDelay > ExploseTime
	      || RedPowerDelay > ExploseTime)
	    {
	      RedPowerDelay -= RedStep;
	      GreenPowerDelay -= GreenStep;
	      BluePowerDelay -= BlueStep;
	      break;
	    }
	}
      else
	{
	  RedPowerDelay -= RedStep;
	  GreenPowerDelay -= GreenStep;
	  BluePowerDelay -= BlueStep;
	  if (BluePowerDelay < 0.0 || GreenPowerDelay < 0.0
	      || RedPowerDelay < 0.0)
	    {
	      RedPowerDelay += RedStep;	/* out of boundary */
	      GreenPowerDelay += GreenStep;
	      BluePowerDelay += BlueStep;
	      break;
	    }
	}
    }
  if (ssCurrent == SG_BRIGHTER && ssLast == SG_DARKER)
    {
      RedPowerDelay += RedStep;
      GreenPowerDelay += GreenStep;
      BluePowerDelay += BlueStep;
    }
  *RedTarget = (SANE_Byte) (RedPowerDelay);
  *BlueTarget = (SANE_Byte) (BluePowerDelay);
  *GreenTarget = (SANE_Byte) (GreenPowerDelay);
  usb_low_SetRedPD (dev->chip, *RedTarget);
  usb_low_SetGreenPD (dev->chip, *BlueTarget);
  usb_low_SetBluePD (dev->chip, *GreenTarget);
}
#endif
void
usb_high_scan_AdjustRgb600PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Byte MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  SIGNALSTATE ssCurrent = SG_UNKNOW;

  if (dev->isAdjustedRgb600PowerDelay)
    return;
  /* Setup Initial State */
  dev->RedRgb600PowerDelay = MaxPowerDelay;
  dev->GreenRgb600PowerDelay = MaxPowerDelay;
  dev->BlueRgb600PowerDelay = MaxPowerDelay;

  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetDummy (dev->chip, dev->INI_600SkipsPerRow);
  usb_low_SetImageByteWidth (dev->chip, A4CIS_600AdjustLength);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);

  /* */
  /* Adjust GreenPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetGreenPGA (dev->chip,
			     dev->GreenRgb600PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetGreenPD, &ssCurrent,
			    &dev->GreenRgb600PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength);

  /* Adjust BluePD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_BLUE);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb600PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetBluePD, &ssCurrent,
			    &dev->BlueRgb600PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength);

  /* Adjust RedPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_RED);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb600PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetRedPD, &ssCurrent,
			    &dev->RedRgb600PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength);

  dev->isAdjustedRgb600PowerDelay = SANE_TRUE;
}

void
usb_high_scan_AdjustMono600PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Byte MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  SIGNALSTATE ssCurrent = SG_UNKNOW;

  if (dev->isAdjustedMono600PowerDelay)
    return;
  /* Setup Initial State */
  dev->RedMono600PowerDelay = MaxPowerDelay;
  dev->GreenMono600PowerDelay = MaxPowerDelay;
  dev->BlueMono600PowerDelay = MaxPowerDelay;

  /* Compute Gray PD */
  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetDummy (dev->chip, dev->INI_600SkipsPerRow);
  usb_low_SetImageByteWidth (dev->chip, A4CIS_600AdjustLength);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);

  /* */
#ifdef GREENGRAY
  /* Adjust GreenGrayPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (dev->chip, dev->Mono600PGA);
  usb_mid_front_SetGreenPGA (dev->chip, dev->Mono600PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->Mono600PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetGreenPD, &ssCurrent,
			    &dev->GreenMono600PowerDelay,
			    MaxPowerDelay, 0, dev->INI_MaxPowerDelay,
			    A4CIS_600AdjustLength);
#else
  /* Adjust GreenGrayPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetGreenPGA (dev->chip, dev->Mono600PGA);
  usb_high_scan_BSSCPowerDelay (dev->chip, &usb_low_SetGreenPD,
			    ssCurrent, dev->GreenMono600PowerDelay,
			    MaxPowerDelay, 0, dev->INI_MaxPowerDelay,
			    A4CIS_600AdjustLength);

  /* Adjust BluePD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_BLUE);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetBluePGA (dev->chip, dev->Mono600PGA);
  usb_high_scan_BSSCPowerDelay (dev->chip, &usb_low_SetBluePD,
			    ssCurrent, dev->BlueMono600PowerDelay,
			    MaxPowerDelay, 0, dev->INI_MaxPowerDelay,
			    A4CIS_600AdjustLength);

  /* Adjust RedPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_RED);
  /* Modified by Eden */
  if (dev->Is300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 600);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 600);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (dev->chip, dev->Mono600PGA);
  usb_high_scan_BSSCPowerDelay (dev->chip, &usb_low_SetRedPD,
			    ssCurrent, dev->RedMono600PowerDelay,
			    MaxPowerDelay, 0, dev->INI_MaxPowerDelay,
			    A4CIS_600AdjustLength);

  /* Adjust AllPD */
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareMonoX (dev->chip, 600);
  else
    usb_mid_c600_PrepareMonoX (dev->chip, 600);
  /* */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);

  usb_mid_front_SetRedPGA (dev->chip, Mono600PGA);
  usb_mid_front_SetGreenPGA (dev->chip, Mono600PGA);
  usb_mid_front_SetBluePGA (dev->chip, Mono600PGA);
  usb_high_scan_LSACPowerDelay (dev->chip, RedMono600PowerDelay,
			    GreenMono600PowerDelay, BlueMono600PowerDelay,
			    dev->INI_MaxPowerDelay, A4CIS_600AdjustLength);
#endif
  dev->isAdjustedMono600PowerDelay = SANE_TRUE;
}

void
usb_high_scan_AdjustRgb300PowerDelay (Mustek_Usb_Device * dev)
{
  /* Setup Initial State */
  SANE_Byte MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  SIGNALSTATE ssCurrent = SG_UNKNOW;

  if (dev->isAdjustedRgb300PowerDelay)
    return;

  dev->RedRgb300PowerDelay = MaxPowerDelay;
  dev->GreenRgb300PowerDelay = MaxPowerDelay;
  dev->BlueRgb300PowerDelay = MaxPowerDelay;

  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetDummy (dev->chip, dev->INI_300SkipsPerRow);
  usb_low_SetImageByteWidth (dev->chip, A4CIS_300AdjustLength);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);

  /* Adjust GreenPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 300);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 300);
  /* */

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetGreenPGA (dev->chip,
			     dev->GreenRgb300PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetGreenPD, &ssCurrent,
			    &dev->GreenRgb300PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);

  /* Adjust BluePD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_BLUE);
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 300);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 300);
  /* */

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb300PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetBluePD, &ssCurrent,
			    &dev->BlueRgb300PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);

  /* Adjust RedPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_RED);
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 300);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 300);

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb300PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetRedPD, &ssCurrent,
			    &dev->RedRgb300PowerDelay, MaxPowerDelay,
			    0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);
  dev->isAdjustedRgb300PowerDelay = SANE_TRUE;
}

void
usb_high_scan_AdjustMono300PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Byte MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  SIGNALSTATE ssCurrent = SG_UNKNOW;

  if (dev->isAdjustedMono300PowerDelay)
    return;
  /* Setup Initial State */
  dev->RedMono300PowerDelay = MaxPowerDelay;
  dev->GreenMono300PowerDelay = MaxPowerDelay;
  dev->BlueMono300PowerDelay = MaxPowerDelay;

  /* Compute Gray PD */
  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetDummy (dev->chip, dev->INI_300SkipsPerRow);
  usb_low_SetImageByteWidth (dev->chip, A4CIS_300AdjustLength);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);

#ifdef GREENGRAY
  /* Adjust GreenGrayPD */
  usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN);

  /* Modified by Eden */
  if (dev->bIs300600Switch)
    usb_mid_c300600_PrepareRgbX (dev->chip, 300);
  else
    usb_mid_c600_PrepareRgbX (dev->chip, 300);
  /* */

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (dev->chip, dev->Mono300PGA);
  usb_mid_front_SetGreenPGA (dev->chip, dev->Mono300PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->Mono300PGA);
  usb_high_scan_BSSCPowerDelay (dev, &usb_low_SetGreenPD, &ssCurrent,
			    &dev->GreenMono300PowerDelay,
			    MaxPowerDelay, 0, dev->INI_MaxPowerDelay,
			    A4CIS_300AdjustLength);
#else
  /* Adjust GreenGrayPD */
  Motor.PrepareAdjust (CH_GREEN);
  /* Modified by Eden */
  if (bIs300600Switch)
    Sensor300600.PrepareRgbX (300);
  else
    Sensor600.PrepareRgbX (300);
  /* */

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetGreenPGA (Mono300PGA);
  BSSCPowerDelay (&usb_low_SetGreenPD, ssCurrent, GreenMono300PowerDelay,
		  MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);

  /* Adjust BluePD */
  Motor.PrepareAdjust (CH_BLUE);
  /* Modified by Eden */
  if (bIs300600Switch)
    Sensor300600.PrepareRgbX (300);
  else
    Sensor600.PrepareRgbX (300);
  /* */
  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetBluePGA (Mono300PGA);
  BSSCPowerDelay (&usb_low_SetBluePD, ssCurrent, BlueMono300PowerDelay,
		  MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);

  /* Adjust RedPD */
  Motor.PrepareAdjust (CH_RED);
  /* Modified by Eden */
  if (bIs300600Switch)
    Sensor300600.PrepareRgbX (300);
  else
    Sensor600.PrepareRgbX (300);
  /* */

  ssCurrent = SG_UNKNOW;
  usb_mid_front_SetRedPGA (Mono300PGA);
  BSSCPowerDelay (&usb_low_SetRedPD, ssCurrent, RedMono300PowerDelay,
		  MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength);

  /* Adjust AllPD */
  /* Modified by Eden */
  if (bIs300600Switch)
    Sensor300600.PrepareMonoX (300);
  else
    Sensor600.PrepareRgbX (300);

  Motor.PrepareAdjust (CH_GREEN);

  usb_mid_front_SetRedPGA (Mono300PGA);
  usb_mid_front_SetGreenPGA (Mono300PGA);
  usb_mid_front_SetBluePGA (Mono300PGA);
  LSACPowerDelay (RedMono300PowerDelay, GreenMono300PowerDelay,
		  BlueMono300PowerDelay, dev->INI_MaxPowerDelay,
		  A4CIS_300AdjustLength);
#endif
  dev->isAdjustedMono300PowerDelay = SANE_TRUE;
}

void
usb_high_scan_EvaluePixelRate (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_EvaluePixelRate: dev=%p\n", dev);
  dev->PixelRate = 2000;
  dev->isEvaluePixelRate = SANE_TRUE;
}

#if 0
SANE_Word
usb_high_scan_FiltLarger (Mustek_Usb_Device * dev, SANE_Word * pSort,
			  SANE_Word TotalCount, SANE_Word FiltCount)
{
  SANE_Word Bound = TotalCount - 1;
  /* SANE_Word LeftCount=TotalCount-FiltCount; */
  int Temp = 0;
  SANE_Word i, j;

  DBG (5, "usb_high_scan_FiltLarger: dev=%p\n", dev);

  for (i = 0; i < Bound; i++)
    {
      for (j = 0; j < Bound - i; j++)
	{
	  if (pSort[j + 1] < pSort[j])
	    {
	      Temp = pSort[j];
	      pSort[j] = pSort[j + 1];
	      pSort[j + 1] = Temp;
	    }
	}
    }
  return pSort[FiltCount];
}

double
usb_high_scan_FiltLower (Mustek_Usb_Device * dev, int *pSort, SANE_Word TotalCount,
		     SANE_Word FiltCount)
{
  SANE_Word Bound = TotalCount - 1;
  SANE_Word LeftCount = TotalCount - FiltCount;
  int Temp = 0;
  SANE_Word i, j;
  int Sum;

  DBG (5, "usb_high_scan_FiltLower: dev=%p\n", dev);

  for (i = 0; i < Bound; i++)
    {
      for (j = 0; j < Bound - i; j++)
	{
	  if (pSort[j + 1] > pSort[j])
	    {
	      Temp = pSort[j];
	      pSort[j] = pSort[j + 1];
	      pSort[j + 1] = Temp;
	    }
	}
    }
  Sum = 0;
  for (i = 0; i < LeftCount; i++)
    Sum += pSort[i];
  return (double) (Sum);
}
#endif

void
usb_high_scan_CalibrationRgb48Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_CalibrationRgb48Bit: dev=%p\n", dev);
  
/*
	try {
		CURMOTOR Motor(MA1017);
		Motor.PrepareCalibrateRgbY(YDpi);
		usb_low_TurnLampPower(SANE_TRUE);

		SANE_Word WhiteNeed;
		SANE_Word DarkNeed;
		pRedCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pRedCalibrator->Init(dev->INI_KLevel<<8,0);
		pRedCalibrator->Prepare(Width);
		pRedCalibrator->EmbedGamma(pGammaTable);
		pRedCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);
		pGreenCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pGreenCalibrator->Init(dev->INI_KLevel<<8,0);
		pGreenCalibrator->Prepare(Width);
		pGreenCalibrator->EmbedGamma(pGammaTable+((pGammaTable==NULL)? 0:65536));
		pGreenCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);
		pBlueCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pBlueCalibrator->Init(dev->INI_KLevel<<8,0);
		pBlueCalibrator->Prepare(Width);
		pBlueCalibrator->EmbedGamma(pGammaTable+((pGammaTable==NULL)? 0:131072));
		pBlueCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);

		SANE_Byte * pRed=pRed+SkipsPerRow;
		SANE_Byte * pGreen=pGreen+SkipsPerRow;
		SANE_Byte * pBlue=pBlue+SkipsPerRow;
		//K White
		usb_low_StartRowing();
		SANE_Word i;
		for (i=0;i<WhiteNeed;i++){
			SANE_Word j;
			//Read Green Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pGreen);
				pGreenCalibrator->FillInWhite(i,j,(void *)(pGreen));
			}
			//Read Blue Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pBlue);
				pBlueCalibrator->FillInWhite(i,j,(void *)(pBlue));
			}
			//Read Red Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pRed);
				pRedCalibrator->FillInWhite(i,j,(void *)(pRed));
			}
		}
		usb_low_StopRowing();
		// Caculate Aver
		pGreenCalibrator->EvalueWhite(dev->INI_GreenFactor);
		pBlueCalibrator->EvalueWhite(dev->INI_BlueFactor);
		pRedCalibrator->EvalueWhite(dev->INI_RedFactor);

		Motor.PrepareCalibrateRgbY(YDpi);
		usb_low_EnableMotor(SANE_FALSE);
		usb_low_TurnLampPower(SANE_FALSE);

		//K Black
		usb_low_StartRowing();
		for (i=0;i<DarkNeed;i++){
			SANE_Word j;
			//Read Green Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pGreen);
				pGreenCalibrator->FillInDark(i,j,(void *)(pGreen));
			}
			//Read Blue Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pBlue);
				pBlueCalibrator->FillInDark(i,j,(void *)(pBlue));
			}
			//Read Red Channel
			for (j=0;j<2;j++) {
				usb_low_GetRow(pRed);
				pRedCalibrator->FillInDark(i,j,(void *)(pRed));
			}
		}
		usb_low_StopRowing();
		usb_low_TurnLampPower(SANE_TRUE);
		//Caculate Aver
		pGreenCalibrator->EvalueDark(dev->INI_GreenBlackFactor);
		pBlueCalibrator->EvalueDark(dev->INI_BlueBlackFactor);
		pRedCalibrator->EvalueDark(dev->INI_RedBlackFactor);
		// Caculate Mapping
		pGreenCalibrator->EvalueCalibrator();
		pBlueCalibrator->EvalueCalibrator();
		pRedCalibrator->EvalueCalibrator();
	}
	catch (XMA1017 x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_CalibrationRgb24Bit Exception Occured.");
	}
	catch (XCalibrator x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_CalibrationRgb24Bit XCalibrator x:%s",(SANE_String_Const)x);
	}
*/
}

void
usb_high_scan_CalibrationRgb36Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_CalibrationRgb36Bit: dev=%p\n", dev);
/*
	try {
		CURMOTOR Motor(MA1017);
		Motor.PrepareCalibrateRgbY(YDpi);
		usb_low_TurnLampPower(SANE_TRUE);

		SANE_Word WhiteNeed;
		SANE_Word DarkNeed;
		pRedCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pRedCalibrator->Init(dev->INI_KLevel<<8,0);
		pRedCalibrator->Prepare(Width);
		pRedCalibrator->EmbedGamma(pGammaTable);
		pRedCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);
		pGreenCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pGreenCalibrator->Init(dev->INI_KLevel<<8,0);
		pGreenCalibrator->Prepare(Width);
		pGreenCalibrator->EmbedGamma(pGammaTable+((pGammaTable==NULL)? 0:65536));
		pGreenCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);
		pBlueCalibrator=(ICalibrator*) new CCalibratorMD16i12o16Rgb;
		pBlueCalibrator->Init(dev->INI_KLevel<<8,0);
		pBlueCalibrator->Prepare(Width);
		pBlueCalibrator->EmbedGamma(pGammaTable+((pGammaTable==NULL)? 0:131072));
		pBlueCalibrator->Setup(1,2,8,Width,WhiteNeed,DarkNeed);

		//K White
		usb_low_StartRowing();
		SANE_Word i;
		for (i=0;i<WhiteNeed;i++){
			//Read Green Channel
			usb_low_GetRow(pGreen);
			pGreenCalibrator->FillInWhite(i,0,(void *)(pGreen+SkipsPerRow));
			usb_low_GetRow(pGreen);
			pGreenCalibrator->FillInWhite(i,1,(void *)(pGreen+SkipsPerRow));
			//Read Blue Channel
			usb_low_GetRow(pBlue);
			pBlueCalibrator->FillInWhite(i,0,(void *)(pBlue+SkipsPerRow));
			usb_low_GetRow(pBlue);
			pBlueCalibrator->FillInWhite(i,1,(void *)(pBlue+SkipsPerRow));
			//Read Red Channel
			usb_low_GetRow(pRed);
			pRedCalibrator->FillInWhite(i,0,(void *)(pRed+SkipsPerRow));
			usb_low_GetRow(pRed);
			pRedCalibrator->FillInWhite(i,1,(void *)(pRed+SkipsPerRow));
		}
		usb_low_StopRowing();
		// Caculate Aver
		pGreenCalibrator->EvalueWhite(dev->INI_GreenFactor);
		pBlueCalibrator->EvalueWhite(dev->INI_BlueFactor);
		pRedCalibrator->EvalueWhite(dev->INI_RedFactor);

		Motor.PrepareCalibrateRgbY(YDpi);
		usb_low_EnableMotor(SANE_FALSE);
		usb_low_TurnLampPower(SANE_FALSE);

		//K Black
		usb_low_StartRowing();
		for (i=0;i<DarkNeed;i++){
			//Read Green Channel
			usb_low_GetRow(pGreen);
			pGreenCalibrator->FillInDark(i,0,(void *)(pGreen+SkipsPerRow));
			usb_low_GetRow(pGreen);
			pGreenCalibrator->FillInDark(i,1,(void *)(pGreen+SkipsPerRow));
			//Read Blue Channel
			usb_low_GetRow(pBlue);
			pBlueCalibrator->FillInDark(i,0,(void *)(pBlue+SkipsPerRow));
			usb_low_GetRow(pBlue);
			pBlueCalibrator->FillInDark(i,1,(void *)(pBlue+SkipsPerRow));
			//Read Red Channel
			usb_low_GetRow(pRed);
			pRedCalibrator->FillInDark(i,0,(void *)(pRed+SkipsPerRow));
			usb_low_GetRow(pRed);
			pRedCalibrator->FillInDark(i,1,(void *)(pRed+SkipsPerRow));
		}
		usb_low_StopRowing();
		usb_low_TurnLampPower(SANE_TRUE);
		//Caculate Aver
		pGreenCalibrator->EvalueDark(dev->INI_GreenBlackFactor);
		pBlueCalibrator->EvalueDark(dev->INI_BlueBlackFactor);
		pRedCalibrator->EvalueDark(dev->INI_RedBlackFactor);
		// Caculate Mapping
		pGreenCalibrator->EvalueCalibrator();
		pBlueCalibrator->EvalueCalibrator();
		pRedCalibrator->EvalueCalibrator();
	}
	catch (XMA1017 x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_CalibrationRgb24Bit Exception Occured.");
	}
	catch (XCalibrator x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_CalibrationRgb24Bit XCalibrator x:%s",(SANE_String_Const)x);
	}
*/
}

void
usb_high_scan_CalibrationRgb24Bit (Mustek_Usb_Device * dev)
{
  SANE_Word WhiteNeed;
  SANE_Word DarkNeed;
  SANE_Word i;

  usb_mid_motor1200_PrepareCalibrateRgbY (dev->chip, dev->YDpi);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);

  dev->pRedCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  usb_high_cal_MD12_init (dev->pRedCalibrator, I8O8RGB);
  usb_high_cal_MD12_Init (dev->pRedCalibrator, dev->INI_KLevel << 8, 0);
  usb_high_cal_MD12_Prepare (dev->pRedCalibrator,
			     dev->Width);
  usb_high_cal_MD12_EmbedGamma (dev->pRedCalibrator,
			       dev->pGammaTable);
  usb_high_cal_MD12_Setup (dev->pRedCalibrator, 1, 2, 8,
			  dev->Width, &WhiteNeed, &DarkNeed);

  dev->pGreenCalibrator =
    (CalMD12 *) malloc (sizeof (CalMD12));
  usb_high_cal_MD12_init (dev->pGreenCalibrator, I8O8RGB);
  usb_high_cal_MD12_Init (dev->pGreenCalibrator, dev->INI_KLevel << 8, 0);
  usb_high_cal_MD12_Prepare (dev->pGreenCalibrator, dev->Width);
  usb_high_cal_MD12_EmbedGamma (dev->pGreenCalibrator,
			       dev->pGammaTable);
  usb_high_cal_MD12_Setup (dev->pGreenCalibrator, 1, 2, 8,
			  dev->Width, &WhiteNeed, &DarkNeed);

  dev->pBlueCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  usb_high_cal_MD12_init (dev->pBlueCalibrator, I8O8RGB);
  usb_high_cal_MD12_Init (dev->pBlueCalibrator, dev->INI_KLevel << 8, 0);
  usb_high_cal_MD12_Prepare (dev->pBlueCalibrator, dev->Width);
  usb_high_cal_MD12_EmbedGamma (dev->pBlueCalibrator,
			       dev->pGammaTable);
  usb_high_cal_MD12_Setup (dev->pBlueCalibrator, 1, 2, 8,
			  dev->Width, &WhiteNeed, &DarkNeed);

  /* K White */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < WhiteNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->pGreen);
      usb_high_cal_MD12_FillInWhite (dev->pGreenCalibrator, i, 0,
				    (void *) (dev->pGreen +
					      dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pGreen);
      usb_high_cal_MD12_FillInWhite (dev->pGreenCalibrator, i, 1,
				    (void *) (dev->pGreen +
					      dev->SkipsPerRow));
      /* Read Blue Channel */
      usb_low_GetRow (dev->chip, dev->pBlue);
      usb_high_cal_MD12_FillInWhite (dev->pBlueCalibrator, i, 0,
				    (void *) (dev->pBlue +
					      dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pBlue);
      usb_high_cal_MD12_FillInWhite (dev->pBlueCalibrator, i, 1,
				    (void *) (dev->pBlue +
					      dev->SkipsPerRow));
      /* Read Red Channel */
      usb_low_GetRow (dev->chip, dev->pRed);
      usb_high_cal_MD12_FillInWhite (dev->pRedCalibrator, i, 0,
				    (void *) (dev->pRed +
					      dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pRed);
      usb_high_cal_MD12_FillInWhite (dev->pRedCalibrator, i, 1,
				    (void *) (dev->pRed +
					      dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  /* Caculate Aver */
  usb_high_cal_MD12_EvalueWhite (dev->pGreenCalibrator,
				dev->INI_GreenFactor);
  usb_high_cal_MD12_EvalueWhite (dev->pBlueCalibrator, dev->INI_BlueFactor);
  usb_high_cal_MD12_EvalueWhite (dev->pRedCalibrator, dev->INI_RedFactor);

  usb_mid_motor1200_PrepareCalibrateRgbY (dev->chip,
				       dev->YDpi);
  usb_low_EnableMotor (dev->chip, SANE_FALSE);
  usb_low_TurnLampPower (dev->chip, SANE_FALSE);

  /* K Black */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < DarkNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->pGreen);
      usb_high_cal_MD12_FillInDark (dev->pGreenCalibrator, i, 0,
				   (void *) (dev->pGreen +
					     dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pGreen);
      usb_high_cal_MD12_FillInDark (dev->pGreenCalibrator, i, 1,
				   (void *) (dev->pGreen +
					     dev->SkipsPerRow));
      /* Read Blue Channel */
      usb_low_GetRow (dev->chip, dev->pBlue);
      usb_high_cal_MD12_FillInDark (dev->pBlueCalibrator, i, 0,
				   (void *) (dev->pBlue +
					     dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pBlue);
      usb_high_cal_MD12_FillInDark (dev->pBlueCalibrator, i, 1,
				   (void *) (dev->pBlue +
					     dev->SkipsPerRow));
      /* Read Red Channel */
      usb_low_GetRow (dev->chip, dev->pRed);
      usb_high_cal_MD12_FillInDark (dev->pRedCalibrator, i, 0,
				   (void *) (dev->pRed +
					     dev->SkipsPerRow));
      usb_low_GetRow (dev->chip, dev->pRed);
      usb_high_cal_MD12_FillInDark (dev->pRedCalibrator, i, 1,
				   (void *) (dev->pRed +
					     dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);
  /* Caculate Aver */
  usb_high_cal_MD12_EvalueDark (dev->pGreenCalibrator,
			       dev->INI_GreenBlackFactor);
  usb_high_cal_MD12_EvalueDark (dev->pBlueCalibrator,
			       dev->INI_BlueBlackFactor);
  usb_high_cal_MD12_EvalueDark (dev->pRedCalibrator,
			       dev->INI_RedBlackFactor);
  /* Caculate Mapping */
  usb_high_cal_MD12_EvalueCalibrator (dev->pGreenCalibrator);
  usb_high_cal_MD12_EvalueCalibrator (dev->pBlueCalibrator);
  usb_high_cal_MD12_EvalueCalibrator (dev->pRedCalibrator);
}

void
usb_high_scan_CalibrationMono8Bit (Mustek_Usb_Device * dev)
{
  SANE_Word WhiteNeed;
  SANE_Word DarkNeed;
  SANE_Word i;

  usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip,
					dev->YDpi);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);

  dev->pMonoCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  usb_high_cal_MD12_init (dev->pMonoCalibrator, I8O8MONO);
  usb_high_cal_MD12_Init (dev->pMonoCalibrator, dev->INI_KLevel << 8, 0);
  usb_high_cal_MD12_Prepare (dev->pMonoCalibrator, dev->Width);
  usb_high_cal_MD12_EmbedGamma (dev->pMonoCalibrator,
			       dev->pGammaTable);
  usb_high_cal_MD12_Setup (dev->pMonoCalibrator, 1, 1, 8,
			  dev->Width, &WhiteNeed, &DarkNeed);

  /* K White */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < WhiteNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
      usb_high_cal_MD12_FillInWhite (dev->pMonoCalibrator, i, 0,
				    (void *) (dev->A4CIS_GrayLine +
					      dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  /* Caculate Aver */
  usb_high_cal_MD12_EvalueWhite (dev->pMonoCalibrator, dev->INI_GrayFactor);

  usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip,
					dev->YDpi);
  usb_low_EnableMotor (dev->chip, SANE_FALSE);
  usb_low_TurnLampPower (dev->chip, SANE_FALSE);

  /* K Black */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < DarkNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
      usb_high_cal_MD12_FillInDark (dev->pMonoCalibrator, i, 0,
				   (void *) (dev->A4CIS_GrayLine +
					     dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);
  /* Caculate Green Black */
  usb_high_cal_MD12_EvalueDark (dev->pMonoCalibrator,
			       dev->INI_GrayBlackFactor);
  /* Caculate Mapping */
  usb_high_cal_MD12_EvalueCalibrator (dev->pMonoCalibrator);
}

void
usb_high_scan_CalibrationMono4Bit (Mustek_Usb_Device * dev)
{
  SANE_Word WhiteNeed;
  SANE_Word DarkNeed;
  SANE_Word i;

  usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip,
					dev->YDpi);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);

  dev->pMonoCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  usb_high_cal_MD12_init (dev->pMonoCalibrator, I4O1MONO);
  usb_high_cal_MD12_Init (dev->pMonoCalibrator, dev->INI_KLevel << 8, 0);
  usb_high_cal_MD12_Prepare (dev->pMonoCalibrator, dev->Width);
  usb_high_cal_MD12_SetThreshold (dev->pMonoCalibrator,
				 (SANE_Word) (dev->Threshold) << 8);
  usb_high_cal_MD12_Setup (dev->pMonoCalibrator, 1, 1, 8,
			  dev->Width, &WhiteNeed, &DarkNeed);

  /* K White */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < WhiteNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
      usb_high_cal_MD12_FillInWhite (dev->pMonoCalibrator, i, 0,
				    (void *) (dev->A4CIS_GrayLine +
					      dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  /* Caculate Aver */
  usb_high_cal_MD12_EvalueWhite (dev->pMonoCalibrator, dev->INI_GrayFactor);

  usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip,
					dev->YDpi);
  usb_low_EnableMotor (dev->chip, SANE_FALSE);
  usb_low_TurnLampPower (dev->chip, SANE_FALSE);

  /* K Black */
  usb_low_StartRowing (dev->chip);
  for (i = 0; i < DarkNeed; i++)
    {
      /* Read Green Channel */
      usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
      usb_high_cal_MD12_FillInDark (dev->pMonoCalibrator, i, 0,
				   (void *) (dev->A4CIS_GrayLine +
					     dev->SkipsPerRow));
    }
  usb_low_StopRowing (dev->chip);
  usb_low_TurnLampPower (dev->chip, SANE_TRUE);
  /* Caculate Green Black */
  usb_high_cal_MD12_EvalueDark (dev->pMonoCalibrator,
			       dev->INI_GrayBlackFactor);
  /* Caculate Mapping */
  usb_high_cal_MD12_EvalueCalibrator (dev->pMonoCalibrator);
}

void
usb_high_scan_CalibrationMono1Bit (Mustek_Usb_Device * dev)
{
  usb_high_scan_StepForward (dev, dev->INI_KLines * dev->INI_KLoops);
}

SANE_Bool usb_high_scan_StepForward (Mustek_Usb_Device * dev, int StepCount)
{
  SANE_Byte Byte;

  if (StepCount <= 0)
    return SANE_TRUE;
  /* Initialize */
  usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime);
  usb_low_SetMotorDirection (dev->chip, SANE_FALSE);
  usb_mid_motor1200_PrepareStep (dev->chip, (SANE_Word) StepCount);
  /* Startup */
  usb_low_StartRowing (dev->chip);
  /* Wait to Stop */
  /* Linux USB seems buggy on timeout... sleep before really try  */
  /* to read the flag from scanner */
  usleep (StepCount * 2 * 1000);
  Byte = usb_low_WaitRowingStop (dev->chip);
  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);

  return SANE_TRUE;
}

SANE_Bool usb_high_scan_SafeForward (Mustek_Usb_Device * dev, int StepCount)
{
  SANE_Byte Byte;

  if (StepCount <= 0)
    return SANE_TRUE;
  /* Initialize */
  usb_low_SetCCDWidth (dev->chip, 5400);
  usb_low_SetMotorDirection (dev->chip, SANE_FALSE);
  usb_mid_motor1200_PrepareStep (dev->chip, (SANE_Word) StepCount);
  /* Startup */
  usb_low_StartRowing (dev->chip);
  /* Wait to Stop */
  Byte = usb_low_WaitRowingStop (dev->chip);
  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  return SANE_TRUE;
}

#if 0
SANE_Bool usb_high_scan_StepBackward (Mustek_Usb_Device * dev, int StepCount)
{
  SANE_Byte Byte;
  if (StepCount <= 0)
    return SANE_TRUE;
  /* Initialize */
  usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime);
  usb_low_SetMotorDirection (dev->chip, SANE_TRUE);
  usb_mid_motor1200_PrepareStep (dev->chip, (SANE_Word) StepCount);
  /* Startup */
  usb_low_StartRowing (dev->chip);
  Byte = usb_low_WaitRowingStop (dev->chip);
  usb_low_SetCCDWidth (dev->chip, dev->ExploseTime);
  return SANE_TRUE;
}
#endif

#if 0
SANE_Bool
usb_high_scan_StepBackForward (Mustek_Usb_Device * dev, int BackStepCount,
			       int ForStepCount)
{
  DBG (5, "usb_high_scan_StepBackForward: dev=%p\n", dev);
  usb_high_scan_StepBackward (dev, BackStepCount);
  usb_high_scan_StepForward (dev, ForStepCount);
  return SANE_TRUE;
}
#endif

SANE_Word
usb_high_scan_CaculateMaxRgb600Explose (Mustek_Usb_Device * dev,
					SANE_Byte * IdealRedPD,
					SANE_Byte * IdealGreenPD,
					SANE_Byte * IdealBluePD)
{
  SANE_Word RedLightUp =
    dev->ExploseTime - dev->RedRgb600PowerDelay * 64;
  SANE_Word GreenLightUp =
    dev->ExploseTime - dev->GreenRgb600PowerDelay * 64;
  SANE_Word BlueLightUp =
    dev->ExploseTime - dev->BlueRgb600PowerDelay * 64;
  SANE_Word MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));
  SANE_Word IdealExploseTime =
    max (max (5376, MaxLightUp),
	 usb_mid_motor1200_RgbCapability (dev->YDpi));

  DBG (5, "usb_high_scan_CaculateMaxRgb600Explose: dev=%p\n", dev);

  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxMono600Explose (Mustek_Usb_Device * dev, 
					 SANE_Byte * IdealRedPD,
					 SANE_Byte * IdealGreenPD,
					 SANE_Byte * IdealBluePD)
{
#ifdef GREENGRAY
  SANE_Word MaxLightUp =
    dev->ExploseTime - dev->GreenMono600PowerDelay * 64;
#else
  SANE_Word RedLightUp =
    dev->ExploseTime - dev->RedMono600PowerDelay * 64;
  SANE_Word GreenLightUp =
    dev->ExploseTime - dev->GreenMono600PowerDelay * 64;
  SANE_Word BlueLightUp =
    dev->ExploseTime - dev->BlueMono600PowerDelay * 64;
  SANE_Word MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));
#endif
  SANE_Word IdealExploseTime;
  SANE_Word TransferTime =
    (SANE_Word) ((SANE_Word) (dev->PixelRate) * (SANE_Word) (dev->XDpi) / 600);	/* base on 600, but double it. */

  DBG (5, "usb_high_scan_CaculateMaxMono600Explose: dev=%p\n", dev);

  if (TransferTime > 16000)
    TransferTime = 16000;
  IdealExploseTime =
    max (max (5376, MaxLightUp),
	 max (TransferTime,
	      usb_mid_motor1200_MonoCapability (dev->YDpi)));
  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
#ifdef GREENGRAY
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - MaxLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime) / 64);
#else
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
#endif
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxRgb300Explose (Mustek_Usb_Device * dev, SANE_Byte * IdealRedPD,
				    SANE_Byte * IdealGreenPD, SANE_Byte * IdealBluePD)
{
  SANE_Word RedLightUp =
    dev->ExploseTime - dev->RedRgb300PowerDelay * 64;
  SANE_Word GreenLightUp =
    dev->ExploseTime - dev->GreenRgb300PowerDelay * 64;
  SANE_Word BlueLightUp =
    dev->ExploseTime - dev->BlueRgb300PowerDelay * 64;
  SANE_Word MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));
  /* Modified by Eden  */
  SANE_Word IdealExploseTime;
  if (dev->bIs300600Switch)
    {
      IdealExploseTime =
	max (max (2624, MaxLightUp),
	     usb_mid_motor1200_RgbCapability (dev->YDpi));
    }
  else
    {
      IdealExploseTime =
	max (max (5376, MaxLightUp),
	     usb_mid_motor1200_RgbCapability (dev->YDpi));
    }
  /* */

  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxMono300Explose (Mustek_Usb_Device * dev, SANE_Byte * IdealRedPD,
				     SANE_Byte * IdealGreenPD, SANE_Byte * IdealBluePD)
{
#ifdef GREENGRAY
  SANE_Word MaxLightUp =
    dev->ExploseTime - dev->GreenMono300PowerDelay * 64;
#else
  SANE_Word RedLightUp =
    dev->ExploseTime - dev->RedMono300PowerDelay * 64;
  SANE_Word GreenLightUp =
    dev->ExploseTime - dev->GreenMono300PowerDelay * 64;
  SANE_Word BlueLightUp =
    dev->ExploseTime - dev->BlueMono300PowerDelay * 64;
  SANE_Word MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));
#endif
  SANE_Word TransferTime =
    (SANE_Word) ((SANE_Word) (dev->PixelRate) * (SANE_Word) (dev->XDpi) / 600);	/* base on 600, but double it. */
  SANE_Word IdealExploseTime;

  if (TransferTime > 16000)
    TransferTime = 16000;
  /* Modified by Eden */
  if (dev->bIs300600Switch)
    {
      IdealExploseTime =
	max (max (2688, MaxLightUp),
	     max (TransferTime,
		  usb_mid_motor1200_MonoCapability (dev->YDpi)));
    }
  else
    {
      IdealExploseTime =
	max (max (5376, MaxLightUp),
	     max (TransferTime,
		  usb_mid_motor1200_MonoCapability (dev->YDpi)));
    }
  /* */

  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
#ifdef GREENGRAY
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - MaxLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime) / 64);
#else
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
#endif
  return IdealExploseTime;
}

void
usb_high_scan_PrepareRgbSignal600Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime =
    usb_high_scan_CaculateMaxRgb600Explose (dev, &IdealRedPD, &IdealGreenPD,
					&IdealBluePD);

  usb_low_SetCCDWidth (dev->chip, IdealExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb600PGA);
  usb_mid_front_SetGreenPGA (dev->chip,
			     dev->GreenRgb600PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb600PGA);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetRedPD (dev->chip, IdealRedPD);
  usb_low_SetGreenPD (dev->chip, IdealGreenPD);
  usb_low_SetBluePD (dev->chip, IdealBluePD);
}

void
usb_high_scan_PrepareMonoSignal600Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime =
    usb_high_scan_CaculateMaxMono600Explose (dev, &IdealRedPD, &IdealGreenPD,
					     &IdealBluePD);

  usb_low_SetCCDWidth (dev->chip, IdealExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRedPGA (dev->chip, dev->Mono600PGA);
  usb_mid_front_SetGreenPGA (dev->chip, dev->Mono600PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->Mono600PGA);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetRedPD (dev->chip, IdealRedPD);
  usb_low_SetGreenPD (dev->chip, IdealGreenPD);
  usb_low_SetBluePD (dev->chip, IdealBluePD);
}

void
usb_high_scan_PrepareRgbSignal300Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime =
    usb_high_scan_CaculateMaxRgb300Explose (dev, &IdealRedPD, &IdealGreenPD,
					&IdealBluePD);

  usb_low_SetCCDWidth (dev->chip, IdealExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb300PGA);
  usb_mid_front_SetGreenPGA (dev->chip,
			     dev->GreenRgb300PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb300PGA);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetRedPD (dev->chip, IdealRedPD);
  usb_low_SetGreenPD (dev->chip, IdealGreenPD);
  usb_low_SetBluePD (dev->chip, IdealBluePD);
}

void
usb_high_scan_PrepareMonoSignal300Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime =
    usb_high_scan_CaculateMaxMono300Explose (dev, &IdealRedPD, &IdealGreenPD,
					 &IdealBluePD);

  usb_low_SetCCDWidth (dev->chip, IdealExploseTime);
  usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd);
  usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef);
  usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset);
  usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset);
  usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset);
  usb_mid_front_SetRedPGA (dev->chip, dev->Mono300PGA);
  usb_mid_front_SetGreenPGA (dev->chip, dev->Mono300PGA);
  usb_mid_front_SetBluePGA (dev->chip, dev->Mono300PGA);
  usb_mid_front_SetRGBSignal (dev->chip);
  usb_low_SetRedPD (dev->chip, IdealRedPD);
  usb_low_SetGreenPD (dev->chip, IdealGreenPD);
  usb_low_SetBluePD (dev->chip, IdealBluePD);
}

void
usb_high_scan_PrepareRgb48Bit (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_12BIT);
}

void
usb_high_scan_PrepareRgb36Bit (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_12BIT);
}

void
usb_high_scan_PrepareRgb24Bit (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);
}

void
usb_high_scan_PrepareMono8Bit (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_8BIT);
}

void
usb_high_scan_PrepareMono4Bit (Mustek_Usb_Device * dev)
{
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_4BIT);
}

void
usb_high_scan_PrepareMono1Bit (Mustek_Usb_Device * dev)
{
  usb_low_SetImageByteWidth (dev->chip,
			      dev->BytesPerStrip);
  usb_low_SetDummy (dev->chip, dev->Dummy);
  usb_low_SetPixelDeepth (dev->chip, PD_1BIT);
}

SANE_Bool
usb_high_scan_GetRgb48BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			   SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetRgb48BitLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);
  return SANE_TRUE;

/*
	try {
		usb_low_GetRow(pGreen);
		pGreenCalibrator->Calibrate(pGreen+SkipsPerRow,lpLine+2);

		usb_low_GetRow(pBlue);
		pBlueCalibrator->Calibrate(pBlue+SkipsPerRow,lpLine+((isOrderInvert)?0:4));

		usb_low_GetRow(pRed);
		pRedCalibrator->Calibrate(pRed+SkipsPerRow,lpLine+((isOrderInvert)?4:0));
	}
	catch (XMA1017 x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_GetRgb48BitLine Exception Occure");
	}
	catch (XCalibrator x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_GetRgb48BitLine XCalibrator x:%s",(SANE_String_Const)x);
	}
*/
}

SANE_Bool
usb_high_scan_GetRgb36BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			   SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetRgb36BitLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);
  return SANE_TRUE;
/*
	try {
		usb_low_GetRow(pGreen);
		pGreenCalibrator->Calibrate(pGreen+SkipsPerRow,lpLine+2);

		usb_low_GetRow(pBlue);
		pBlueCalibrator->Calibrate(pBlue+SkipsPerRow,lpLine+((isOrderInvert)?0:4));

		usb_low_GetRow(pRed);
		pRedCalibrator->Calibrate(pRed+SkipsPerRow,lpLine+((isOrderInvert)?4:0));
	}
	catch (XMA1017 x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_GetRgb36BitLine Exception Occure");
	}
	catch (XCalibrator x) {
		x.ReportError();
		throw XScanner(x.ExceptCode(),"usb_high_scan_GetRgb36BitLine XCalibrator x:%s",(SANE_String_Const)x);
	}
*/
}



SANE_Bool
usb_high_scan_GetRgb24BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			       SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetRgb24itLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  usb_low_GetRow (dev->chip, dev->pGreen);

  usb_low_GetRow (dev->chip, dev->pBlue);

  usb_low_GetRow (dev->chip, dev->pRed);
  usb_high_cal_MD12_Calibrate (dev->pGreenCalibrator,
			      dev->pGreen + dev->SkipsPerRow,
			      lpLine + 1);
  usb_high_cal_MD12_Calibrate (dev->pBlueCalibrator,
			      dev->pBlue + dev->SkipsPerRow,
			      lpLine + ((isOrderInvert) ? 0 : 2));
  usb_high_cal_MD12_Calibrate (dev->pRedCalibrator,
			      dev->pRed + dev->SkipsPerRow,
			      lpLine + ((isOrderInvert) ? 2 : 0));

  return SANE_TRUE;
}


SANE_Bool
usb_high_scan_GetMono8BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			   SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetMono8BitLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  if (!usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine))
    {
      DBG (5,"usb_high_scan_GetMono8BitLine Exception Occure");
      return SANE_FALSE;

    }
  usb_high_cal_MD12_Calibrate (dev->pMonoCalibrator,
			       dev->A4CIS_GrayLine +
			       dev->SkipsPerRow,
			       lpLine);
  return SANE_TRUE;
}

SANE_Bool
usb_high_scan_GetMono4BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			   SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetMono4BitLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
  usb_high_cal_MD12_Calibrate (dev->pMonoCalibrator,
			       dev->A4CIS_GrayLine +
			       dev->SkipsPerRow,
			       lpLine);
  return SANE_TRUE;
}

SANE_Bool
usb_high_scan_GetMono1BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			   SANE_Bool isOrderInvert)
{
  DBG (5, "usb_high_scan_GetMono1BitLine: dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
  memcpy (lpLine, dev->A4CIS_GrayLine + dev->SkipsPerRow,
	  dev->BytesPerRow);
  return SANE_TRUE;
}

void
usb_high_scan_BackTrackRgb48Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_BackTrackRgb48Bit: dev=%p\n", dev);
}

void
usb_high_scan_BackTrackRgb36Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_BackTrackRgb36Bit: dev=%p\n", dev);
}

void
usb_high_scan_BackTrackRgb24Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_BackTrackRgb24Bit: dev=%p\n", dev);
}

void
usb_high_scan_BackTrackMono8Bit (Mustek_Usb_Device * dev)
{
  int i;

  DBG (5, "usb_high_scan_BackTrackMono8Bit: dev=%p\n", dev);

  if (dev->YDpi >= 300)
    {
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_TRUE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono8BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
      usleep (100 * 1000);
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_FALSE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono8BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
    }
}

void
usb_high_scan_BackTrackMono4Bit (Mustek_Usb_Device * dev)
{
  int i;
  if (dev->YDpi >= 300)
    {
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_TRUE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono4BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_FALSE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono4BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
    }
}

void
usb_high_scan_BackTrackMono1Bit (Mustek_Usb_Device * dev)
{
  int i;
  if (dev->YDpi >= 300)
    {
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_TRUE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono1BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
      usb_low_StopRowing (dev->chip);
      usb_low_SetMotorDirection (dev->chip, SANE_FALSE);
      usb_low_StartRowing (dev->chip);
      for (i = 0; i < dev->INI_Mono1BitBackTrack; i++)
	{
	  usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine);
	}
    }
}
